import copy
import os
import logging
import warnings
from functools import partial
import hashlib
from qtpy.QtWidgets import QPushButton, QMenu, QAction, QMessageBox, QInputDialog, QLineEdit, QWidget
from qtpy.QtGui import QCursor, QIcon, QMouseEvent
from qtpy.QtCore import Slot, Property, Qt, QSize, QPoint
from .base import PyDMWidget, only_if_channel_set
from ..utilities import IconFont, find_file, is_pydm_app
from ..utilities.macro import parse_macro_string
from ..utilities.stylesheet import merge_widget_stylesheet
from ..display import (load_file, ScreenTarget)
from typing import Optional, List

logger = logging.getLogger(__name__)

_relatedDisplayRuleProperties = {
    'Text': ['setText', str],
    'Filenames': ['filenames', list]
    }


class PyDMRelatedDisplayButton(QPushButton, PyDMWidget, new_properties=_relatedDisplayRuleProperties):
    """
    A QPushButton capable of opening a new Display at the same of at a
    new window.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget for the related display button
    filename : str, optional
        The file to be opened
    init_channel : str, optional
        The channel to be used by the widget
    """
    # Constants for determining where to open the display.
    EXISTING_WINDOW = 0
    NEW_WINDOW = 1

    def __init__(self,parent: Optional[QWidget] = None, filename: str = None, init_channel: Optional[str] = None) -> None:
        QPushButton.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)

        self.mouseReleaseEvent = self.push_button_release_event
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)

        self.iconFont = IconFont()
        self._icon = self.iconFont.icon("file")
        self.setIconSize(QSize(16, 16))
        self.setIcon(self._icon)

        self._filenames = [filename] if filename is not None else []
        self._titles = []
        self._macros = []
        self.num_additional_items = 0
        self._shift_key_was_down = False
        self.setCursor(QCursor(self._icon.pixmap(16, 16)))
        self._display_menu_items = None
        self._display_filename = ""
        self._macro_string = None
        self._open_in_new_window = False
        self.open_in_new_window_action = QAction("Open in New Window", self)
        self.open_in_new_window_action.triggered.connect(
            self.handle_open_new_window_action)
        self._show_icon = True
        self._menu_needs_rebuild = True

        self._password_protected = False
        self._password = ""
        self._protected_password = ""

        self._follow_symlinks = False

    @only_if_channel_set
    def check_enable_state(self) -> None:
        """
        override parent method, so this widget does not get disabled when the pv disconnects.
        This method adds a Tool Tip with the reason why it is disabled.
        """
        status = self._connected
        tooltip = self.restore_original_tooltip()
        if not status:
            if tooltip != '':
                tooltip += '\n'
            tooltip += "Alarm PV is disconnected."
            tooltip += '\n'
            tooltip += self.get_address()

        self.setToolTip(tooltip)

    @Property('QStringList')
    def filenames(self) -> List[str]:
        return self._filenames
    
    @filenames.setter
    def filenames(self, val : List[str]) -> None:
        self._filenames = val
        self._menu_needs_rebuild = True
        
    @Property('QStringList')
    def titles(self) -> List[str]:
        return self._titles
    
    @titles.setter
    def titles(self, val : List[str]) -> None:
        self._titles = val
        self._menu_needs_rebuild = True

    def _get_items(self):
        """
        Aggregate file entry information.

        Yields
        ------
        item : dict
            Containing filename, title, and macros/empy macros
            Only containing valid entries or nothing

        """
        for i, filename in enumerate(self.filenames):
            if not filename:
                continue
            item = {'filename': filename}
            if i >= len(self.titles):
                item['title'] = filename
            else:
                item['title'] = self.titles[i]
            if i < len(self.macros):
                item['macros'] = self.macros[i]
            else:
                item['macros'] = ""
            yield item

    def _rebuild_menu(self) -> None:
        if not any(self._filenames):
            self._filenames = []
        if not any(self._titles):
            self._titles = []
        if len(self._filenames) == 0:
            self.setEnabled(False)
        if len(self._filenames) <= 1:
            self.setMenu(None)
            self._menu_needs_rebuild = False
            return
        menu = QMenu(self)
        self._assemble_menu(menu, target=None)
        self.setMenu(menu)
        self._menu_needs_rebuild = False

    @Property(bool)
    def showIcon(self) -> bool:
        """
        Whether or not we should show the selected Icon.

        Returns
        -------
        bool
        """
        return self._show_icon

    @showIcon.setter
    def showIcon(self, value: bool) -> None:
        """
        Whether or not we should show the selected Icon.

        Parameters
        ----------
        value : bool
        """
        if self._show_icon != value:
            self._show_icon = value

            if self._show_icon:
                self.setIcon(self._icon)
            else:
                self._icon = self.icon()
                self.setIcon(QIcon())

    @Property(str, designable=False)
    def displayFilename(self) -> str:
        """
        DEPRECATED: use the 'filenames' property.
        This property simply returns the first filename from the 'filenames'
        property.
        The filename to open

        Returns
        -------
        str
        """
        if len(self.filenames) == 0:
            return ""
        return self.filenames[0]

    @displayFilename.setter
    def displayFilename(self, value: str) -> None:
        """
        DEPRECATED: use the 'filenames' property.
        Any value set to this property is appended to the 'filenames'
        property, then 'displayFilename' is cleared.

        Parameters
        ----------
        value : str
        """
        warnings.warn("'PyDMRelatedDisplayButton.displayFilename' is deprecated, "
                      "use 'PyDMRelatedDisplayButton.filenames' instead.")
        if value:
            if value in self.filenames:
                return
            file_list = [value]
            self.filenames = self.filenames + file_list
        self._display_filename = ""
            
    @Property('QStringList')
    def macros(self) -> List[str]:
        """
        The macro substitutions to use when launching the display, in JSON object format.

        Returns
        -------
        list of str
        """
        return self._macros

    @macros.setter
    def macros(self, new_macros: List[str]) -> None:
        """
        The macro substitutions to use when launching the display, in JSON object format.

        Parameters
        ----------
        new_macros : list of str
        """
        #Handle the deprecated form of macros where it was a single string.
        if isinstance(new_macros, str):
            new_macros = [new_macros]
        self._macros = new_macros

    @Property(bool)
    def openInNewWindow(self) -> bool:
        """
        If true, the button will open the display in a new window, rather than in the existing window.

        Returns
        -------
        bool
        """
        return self._open_in_new_window

    @openInNewWindow.setter
    def openInNewWindow(self, open_in_new: bool) -> None:
        """
        If true, the button will open the display in a new window, rather than in the existing window.

        Parameters
        ----------
        open_in_new : bool
        """
        self._open_in_new_window = open_in_new

    @Property(bool)
    def passwordProtected(self) -> bool:
        """
        Whether or not this button is password protected.
        Returns
        -------
        bool
        -------
        """
        return self._password_protected

    @passwordProtected.setter
    def passwordProtected(self, value: bool) -> None:
        """
        Whether or not this button is password protected.
        Parameters
        ----------
        value : bool
        """
        if self._password_protected != value:
            self._password_protected = value

    @Property(str)
    def password(self) -> str:
        """
        Password to be encrypted using SHA256.

        .. warning::
        To avoid issues exposing the password this method
        always returns an empty string.

        Returns
        -------
        str
        """
        return ""

    @password.setter
    def password(self, value: str) -> None:
        """
        Password to be encrypted using SHA256.

        Parameters
        ----------
        value : str
        The password to be encrypted
        """
        if value is not None and value != "":
            sha = hashlib.sha256()
            sha.update(value.encode())
            # Use the setter as it also checks whether the existing password is the same with the
            # new one, and only updates if the new password is different
            self.protectedPassword = sha.hexdigest()

    @Property(str)
    def protectedPassword(self) -> str:
        """
        The encrypted password.

        Returns
        -------
        str
        """
        return self._protected_password

    @protectedPassword.setter
    def protectedPassword(self, value: str) -> None:
        if self._protected_password != value:
            self._protected_password = value

    @Property(bool)
    def followSymlinks(self) -> bool:
        """
        If True, any symlinks in the path to filename (including the base path of the parent display) will be followed, so that it
        will always use the canonical path. If False (default), the file will be searched without canonicalizing the path beforehand.

        Note that it will not work on Windows if you're using a Python version prior to 3.8.

        Returns
        -------
        bool
        """
        return self._follow_symlinks

    @followSymlinks.setter
    def followSymlinks(self, follow_symlinks: bool) -> None:
        """
        If True, any symlinks in the path to filename (including the base path of the parent display) will be followed, so that it
        will always use the canonical path. If False (default), the file will be searched using the non-canonical path.

        Note that it will not work on Windows if you're using a Python version prior to 3.8.

        Parameters
        ----------
        follow_symlinks : bool
        """
        self._follow_symlinks = follow_symlinks

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self._menu_needs_rebuild:
            self._rebuild_menu()
        if event.button() == Qt.LeftButton and event.modifiers() == Qt.ShiftModifier:
            self._shift_key_was_down = True
        else:
            self._shift_key_was_down = False
        super(PyDMRelatedDisplayButton, self).mousePressEvent(event)

    def push_button_release_event(self, mouse_event: QMouseEvent) -> None:
        """
        Opens the related display given by `filename`.
        If the Shift Key is hold it will open in a new window.

        Called when a mouse button is released. A widget receives mouse
        release events when it has received the corresponding mouse press
        event. This means that if the user presses the mouse inside your
        widget, then drags the mouse somewhere else before releasing the
        mouse button, your widget receives the release event.

        Parameters
        ----------
        mouse_event : QMouseEvent

        """
        if mouse_event.button() != Qt.LeftButton:
            return super(PyDMRelatedDisplayButton, self).mouseReleaseEvent(mouse_event)
        if self.menu() is not None:
            return super(PyDMRelatedDisplayButton, self).mouseReleaseEvent(mouse_event)
        try:
            for item in self._get_items():
                self.open_display(item['filename'], item['macros'],
                                  target=None)
                break
        except Exception:
            logger.exception("Failed to open display.")
        finally:
            super(PyDMRelatedDisplayButton, self).mouseReleaseEvent(
                mouse_event)

    def validate_password(self) -> bool:
        """
        If the widget is ```passwordProtected```, this method will propmt
        the user for the correct password.

        Returns
        -------
        bool
            True in case the password was correct of if the widget is not
            password protected.
        """
        if not self._password_protected:
            return True

        pwd, ok = QInputDialog().getText(None, "Authentication", "Please enter your password:",
                                         QLineEdit.Password, "")
        pwd = str(pwd)
        if not ok or pwd == "":
            return False

        sha = hashlib.sha256()
        sha.update(pwd.encode())
        pwd_encrypted = sha.hexdigest()
        if pwd_encrypted != self._protected_password:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Invalid password.")
            msg.setWindowTitle("Error")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.setDefaultButton(QMessageBox.Ok)
            msg.setEscapeButton(QMessageBox.Ok)
            msg.exec_()
            return False
        return True

    @Slot()
    def handle_open_new_window_action(self) -> None:
        """
        Handle the "Open in New Window" action.

        Returns
        -------
        None.

        """
        for item in self._get_items():
            try:
                self.open_display(item['filename'], item['macros'],
                                  target=self.NEW_WINDOW)
            except Exception:
                logger.exception("Failed to open display.")

    def _assemble_menu(self, menu, target=None):
        for item in self._get_items():
            try:
                action = menu.addAction(item['title'])
                action.triggered.connect(
                    partial(self.open_display, item['filename'],
                            item['macros'], target=target))
            except Exception:
                logger.exception("Failed to open display.")

    @Slot()
    def open_display(self, filename, macro_string="", target=None):
        """
        Open the configured `filename` with the given `target`.

        Parameters
        ----------
        target : int
            PyDMRelatedDisplayButton.EXISTING_WINDOW or 0 will open the
            file on the same window. PyDMRelatedDisplayButton.NEW_WINDOW
            or 1 will result on a new process.

        Returns
        -------
        display : Display
            The widget that was opened. Useful for testing and debug.
        """
        if not self.validate_password():
            return None

        parent_display = self.find_parent_display()
        base_path = ""
        macros = {}
        if parent_display:
            parent_file_path = parent_display.loaded_file()
            if self._follow_symlinks:
                parent_file_path = os.path.realpath(parent_file_path)
            base_path = os.path.dirname(parent_file_path)
            macros = copy.copy(parent_display.macros())

        fname = find_file(filename, base_path=base_path, raise_if_not_found=True)
        widget_macros = parse_macro_string(macro_string)
        macros.update(widget_macros)

        screen_target = None
        if target is self.NEW_WINDOW:
            screen_target = ScreenTarget.NEW_PROCESS
        if self._shift_key_was_down:
            target = self.NEW_WINDOW
            screen_target = ScreenTarget.NEW_PROCESS
        if target is None:
            if self._open_in_new_window:
                target = self.NEW_WINDOW
                screen_target = ScreenTarget.NEW_PROCESS
            else:
                target = self.EXISTING_WINDOW
                screen_target = None

        if is_pydm_app():
            if target == self.NEW_WINDOW:
                return load_file(fname, macros=macros, target=screen_target)
            else:
                return self.window().open(fname, macros=macros)
        else:
            display = load_file(fname, macros=macros, target=ScreenTarget.DIALOG)
            # Not a pydm app: need to give our new display proper pydm styling
            # Usually done in PyDMApplication
            merge_widget_stylesheet(widget=display)
            return display

    def context_menu(self):
        try:
            menu = super(PyDMRelatedDisplayButton, self).context_menu()
        except:
            menu = QMenu(self)
        if len(menu.findChildren(QAction)) > 0:
            menu.addSeparator()
        if len(self.filenames) <= 1:
            menu.addAction(self.open_in_new_window_action)
            return menu
        sub_menu = menu.addMenu("Open in New Window")
        self._assemble_menu(sub_menu, target=self.NEW_WINDOW)
        return menu

    @Slot(QPoint)
    def show_context_menu(self, pos: QPoint) -> None:
        menu = self.context_menu()
        menu.exec_(self.mapToGlobal(pos))




import time
import json
from collections import OrderedDict
from typing import Optional
from pyqtgraph import BarGraphItem, ViewBox, AxisItem
import numpy as np
from qtpy.QtGui import QColor
from qtpy.QtCore import Signal, Slot, Property, QTimer, Q_ENUMS
from .baseplot import BasePlot, BasePlotCurveItem
from .channel import PyDMChannel
from .. utilities import remove_protocol

import logging
logger = logging.getLogger(__name__)

MINIMUM_BUFFER_SIZE = 2
DEFAULT_BUFFER_SIZE = 18000

DEFAULT_X_MIN = -30
DEFAULT_Y_MIN = 0

DEFAULT_TIME_SPAN = 5.0
DEFAULT_UPDATE_INTERVAL = 1000  # Plot update rate for fixed rate mode in milliseconds


class updateMode(object):
    """ updateMode as new type for plot update """
    OnValueChange = 1
    AtFixedRate = 2


class TimePlotCurveItem(BasePlotCurveItem):
    """
    TimePlotCurveItem represents a single curve in a time plot.

    It is used to plot a scalar value vs. time.  In addition to the parameters
    listed below, TimePlotCurveItem accepts keyword arguments for all plot
    options that pyqtgraph.PlotDataItem accepts.

    Parameters
    ----------
    channel_address : str
        The address to of the scalar data to plot.
    plot_by_timestamps : bool
        If True, the x-axis shows timestamps as ticks, and those timestamps
        scroll to the left as time progresses.  If False, the x-axis tick marks
        show time relative to the current time.
    plot_style: str, optional
        Currently one of either 'Line' or 'Bar'. Determines how data points for this
        curve will be plotted. Defaults to a line based plot if not set
    color : QColor, optional
        The color used to draw the curve line and the symbols.
    lineStyle: int, optional
        Style of the line connecting the data points.
        Must be a value from the Qt::PenStyle enum
        (see http://doc.qt.io/qt-5/qt.html#PenStyle-enum).
    lineWidth: int, optional
        Width of the line connecting the data points.
    **kws : dict
        Additional parameters supported by pyqtgraph.PlotDataItem,
        like 'symbol' and 'symbolSize'.
    """
    _channels = ('channel',)

    def __init__(self, channel_address=None, plot_by_timestamps=True, plot_style='Line', **kws):
        """
        Parameters
        ----------
        channel_address : str
            The PV address
        plot_by_timestamps : bool
            If True, the x-axis shows timestamps as ticks, and those timestamps
            scroll to the left as time progresses.  If False, the x-axis tick
            marks show time relative to the current time.
        kws : dict
            Additional parameters supported by pyqtgraph.PlotDataItem,
            like 'symbol' and 'symbolSize'.
        """
        channel_address = "" if channel_address is None else channel_address
        if "name" not in kws or not kws["name"]:
            name = remove_protocol(channel_address)
            kws["name"] = name

        # Keep the x-axis moving with latest timestamps as ticks
        self._plot_by_timestamps = plot_by_timestamps

        self.plot_style = plot_style

        self._bufferSize = MINIMUM_BUFFER_SIZE
        self._update_mode = PyDMTimePlot.OnValueChange

        self._min_y_value = None
        self._max_y_value = None

        self.data_buffer = np.zeros((2, self._bufferSize), order='f', dtype=float)
        self.connected = False
        self.points_accumulated = 0
        self.latest_value = None
        self.channel = None
        self.address = channel_address
        super(TimePlotCurveItem, self).__init__(**kws)

    def to_dict(self):
        dic_ = OrderedDict([("channel", self.address),
                            ("plot_style", self.plot_style)])
        dic_.update(super(TimePlotCurveItem, self).to_dict())
        return dic_

    @property
    def address(self):
        if self.channel is None:
            return None
        return self.channel.address

    @address.setter
    def address(self, new_address):
        if new_address is None or len(str(new_address)) < 1:
            self.channel = None
            return
        self.channel = PyDMChannel(address=new_address,
                                   connection_slot=self.connectionStateChanged,
                                   value_slot=self.receiveNewValue)

    @property
    def plotByTimeStamps(self):
        return self._plot_by_timestamps

    @plotByTimeStamps.setter
    def plotByTimeStamps(self, new_value):
        if self._plot_by_timestamps != new_value:
            self._plot_by_timestamps = new_value

    @property
    def minY(self):
        """
        Get the minimum y-value so far in the same plot. This is useful to
        scale the y-axis for a selected curve.

        Returns
        -------
        float
            The minimum y-value collected so far for this current curve.
        """
        return self._min_y_value

    @property
    def maxY(self):
        """
        Get the maximum y-value so far in the same plot. This is useful to
        scale the y-axis for a selected curve.

        Returns
        -------
        float
            The maximum y-value collected so far for this current curve.
        """
        return self._max_y_value

    @Slot(bool)
    def connectionStateChanged(self, connected):
        # Maybe change pen stroke?
        self.connected = connected
        if not self.connected:
            self.latest_value = np.nan

    @Slot(float)
    @Slot(int)
    def receiveNewValue(self, new_value):
        """
        Rotate and fill the data buffer when a new value is available.

        For Synchronous mode, write the new value into the data buffer
        immediately, and increment the accumulated point counter.
        For Asynchronous, write the new value into a temporary (buffered)
        variable, which will be written to the data buffer when asyncUpdate
        is called.

        This method is usually called by a PyDMChannel when it updates.  You
        can call it yourself to inject data into the curve.

        Parameters
        ----------
        new_value : float
            The new y-value.
        """
        self.update_min_max_y_values(new_value)

        if self._update_mode == PyDMTimePlot.OnValueChange:
            self.data_buffer = np.roll(self.data_buffer, -1)
            # The first array row is to record timestamps, when a new value arrives.
            self.data_buffer[0, self._bufferSize - 1] = time.time()
            # The second array row is to record the actual values.
            self.data_buffer[1, self._bufferSize - 1] = new_value

            if self.points_accumulated < self._bufferSize:
                self.points_accumulated += 1
            self.data_changed.emit()
        elif self._update_mode == PyDMTimePlot.AtFixedRated:
            self.latest_value = new_value

    @Slot()
    def asyncUpdate(self):
        """
        Updates the latest data read from the buffered variable into the data
        buffer, together with the timestamp when this happens. Also increments
        the accumulated point counter.
        """
        if self._update_mode != PyDMTimePlot.AtFixedRated:
            return
        self.data_buffer = np.roll(self.data_buffer, -1)
        self.data_buffer[0, self._bufferSize - 1] = time.time()
        self.data_buffer[1, self._bufferSize - 1] = self.latest_value
        if self.points_accumulated < self._bufferSize:
            self.points_accumulated = self.points_accumulated + 1
        self.data_changed.emit()

    def update_min_max_y_values(self, new_value):
        """
        Update the min and max y-value as a new value is available. This is
        useful for auto-scaling to a specific curve.

        Parameters
        ----------
        new_value : float
            The new y-value just available.
        """
        if self._min_y_value is None and self._max_y_value is None:
            self._min_y_value = self._max_y_value = new_value
        elif self._min_y_value > new_value:
            self._min_y_value = new_value
        elif self._max_y_value < new_value:
            self._max_y_value = new_value

    def initialize_buffer(self):
        """
        Initialize the data buffer used to plot the current curve.
        """
        self.points_accumulated = 0

        # If you don't specify dtype=float, you don't have enough
        # resolution for the timestamp data.
        self.data_buffer = np.zeros((2, self._bufferSize),
                                    order='f', dtype=float)
        self.data_buffer[0].fill(time.time())

    def getBufferSize(self):
        return int(self._bufferSize)

    def setBufferSize(self, value):
        if self._bufferSize != int(value):
            self._bufferSize = max(int(value), MINIMUM_BUFFER_SIZE)
            self.initialize_buffer()

    def resetBufferSize(self):
        if self._bufferSize != DEFAULT_BUFFER_SIZE:
            self._bufferSize = DEFAULT_BUFFER_SIZE
            self.initialize_buffer()

    @Slot()
    def redrawCurve(self, min_x: Optional[float] = None, max_x: Optional[float] = None):
        """
        Redraw the curve with the new data.

        If plot by timestamps, plot the x-axis with the timestamps as the ticks.

        On the other hand, if plot by relative time, take the time diff from
        the starting time of the curve, and plot the data to the time diff
        position on the x-axis.

        Parameters
        ----------
        min_x: float, optional
            The minimum timestamp to render when plotting as a bar graph.
        max_x: float, optional
            The maximum timestamp to render when plotting as a bar graph.
        """
        try:
            x = self.data_buffer[0, -self.points_accumulated:].astype(float)
            y = self.data_buffer[1, -self.points_accumulated:].astype(float)

            if not self._plot_by_timestamps:
                x -= time.time()

            if self.plot_style is None or self.plot_style == 'Line':
                self.setData(y=y, x=x)
            elif self.plot_style == 'Bar':
                # In cases where the buffer size is large, we don't want to render 10,000+ bars on every update
                # if only a fraction of those are actually visible. These 2 indices represent the visible range
                # of the plot, and we will only render bars within that range.
                min_index = np.searchsorted(x, min_x)
                max_index = np.searchsorted(x, max_x) + 1
                self._setBarGraphItem(x=x[min_index:max_index], y=y[min_index:max_index])
        except (ZeroDivisionError, OverflowError):
            # Solve an issue with pyqtgraph and initial downsampling
            pass

    def _setBarGraphItem(self, x, y):
        """ Set the plots points to render as bars. No need to call this directly as it will automatically
            be handled by redrawCurve() """
        if self.points_accumulated == 0 or len(x) == 0 or len(y) == 0:
            return

        brushes = np.array([self.color] * len(x))

        if self.threshold_color is not None:
            if self.upper_threshold is not None:
                brushes[np.argwhere(y > self.upper_threshold)] = self.threshold_color
            if self.lower_threshold is not None:
                brushes[np.argwhere(y < self.lower_threshold)] = self.threshold_color

        self.bar_graph_item.setOpts(x=x, height=y, brushes=brushes)

    def setUpdatesAsynchronously(self, value):
        """
        Check if value is from updatesAsynchronously(bool) or updateMode(int)
        """
        if type(value)==int and value == updateMode.AtFixedRate or type(value)==bool and value is True:
            self._update_mode = PyDMTimePlot.AtFixedRate
        else:
            self._update_mode = PyDMTimePlot.OnValueChange
        self.initialize_buffer()
        
    def resetUpdatesAsynchronously(self):
        self._update_mode = PyDMTimePlot.OnValueChange
        self.initialize_buffer()

    def min_x(self):
        """
        Provide the the oldest valid timestamp from the data buffer.

        Returns
        -------
        float
            The timestamp of the most recent data point recorded into the data buffer.
        """
        return self.data_buffer[0, -self.points_accumulated]

    def max_x(self):
        """
        Provide the the most recent timestamp accumulated from the data buffer.
        This is useful for scaling the x-axis.

        Returns
        -------
        float
            The timestamp of the most recent data point recorded into the data buffer.
        """
        return self.data_buffer[0, -1]

    def channels(self):
        return [self.channel]


class PyDMTimePlot(BasePlot,updateMode):
    """
    PyDMTimePlot is a widget to plot one or more channels vs. time.

    Each curve can plot either a Y-axis waveform vs. its indices,
    or a Y-axis waveform against an X-axis waveform.

    Parameters
    ----------
    parent : optional
        The parent of this widget.
    init_y_channels : list
        A list of scalar channels to plot vs time.
    plot_by_timestamps : bool
        If True, the x-axis shows timestamps as ticks, and those timestamps
        scroll to the left as time progresses.  If False, the x-axis tick marks
        show time relative to the current time.
    background: optional
        The background color for the plot.  Accepts any arguments that
        pyqtgraph.mkColor will accept.
    bottom_axis: AxisItem, optional
        Will set the bottom axis of this plot to the input axis. If not set, will default
        to either a TimeAxisItem if plot_by_timestamps is true, or a regular AxisItem otherwise
    """
    OnValueChange = 1
    AtFixedRated = 2

    Q_ENUMS(updateMode)
    updateMode = updateMode

    plot_redrawn_signal = Signal(TimePlotCurveItem)

    def __init__(self, parent=None, init_y_channels=[], plot_by_timestamps=True, background='default', bottom_axis=None):
        """
        Parameters
        ----------

        parent : Widget
            The parent widget of the chart.
        init_y_channels : list
            A list of scalar channels to plot vs time.
        plot_by_timestamps : bool
            If True, the x-axis shows timestamps as ticks, and those timestamps
            scroll to the left as time progresses.  If False, the x-axis tick
            marks show time relative to the current time.
        background : str, optional
            The background color for the plot.  Accepts any arguments that
            pyqtgraph.mkColor will accept.
        """
        self._updateMode = updateMode.OnValueChange
        self._plot_by_timestamps = plot_by_timestamps

        if bottom_axis is not None:
            self._bottom_axis = bottom_axis
        elif plot_by_timestamps:
            self._bottom_axis = TimeAxisItem('bottom')
        else:
            self.starting_epoch_time = time.time()
            self._bottom_axis = AxisItem('bottom')

        super(PyDMTimePlot, self).__init__(parent=parent, background=background,
                                           axisItems={"bottom": self._bottom_axis})

        # Removing the downsampling while PR 763 is not merged at pyqtgraph
        # Reference: https://github.com/pyqtgraph/pyqtgraph/pull/763
        # self.setDownsampling(ds=True, auto=True, mode="mean")

        if self._plot_by_timestamps:
            self.plotItem.disableAutoRange(ViewBox.XAxis)
            self.getViewBox().setMouseEnabled(x=False)
        else:
            self.plotItem.setRange(xRange=[DEFAULT_X_MIN, 0], padding=0)
            self.plotItem.setLimits(xMax=0)

        self._bufferSize = DEFAULT_BUFFER_SIZE

        self._time_span = DEFAULT_TIME_SPAN  # This is in seconds
        self._update_interval = DEFAULT_UPDATE_INTERVAL

        self.update_timer = QTimer(self)
        self.update_timer.setInterval(self._update_interval)
        self._update_mode = PyDMTimePlot.OnValueChange
        self._needs_redraw = True

        self.labels = {
            "left": None,
            "right": None,
            "bottom": None
        }

        self.units = {
            "left": None,
            "right": None,
            "bottom": None
        }

        for channel in init_y_channels:
            self.addYChannel(channel)

    def initialize_for_designer(self):
        # If we are in Qt Designer, don't update the plot continuously.
        # This function gets called by PyDMTimePlot's designer plugin.
        self.redraw_timer.setSingleShot(True)

    def addYChannel(self, y_channel=None, plot_style=None, name=None, color=None,
                    lineStyle=None, lineWidth=None, symbol=None, symbolSize=None,
                    barWidth=None, upperThreshold=None, lowerThreshold=None, thresholdColor=None,
                    yAxisName=None, useArchiveData=False):
        """
        Adds a new curve to the current plot

        Parameters
        ----------
        y_channel : str
            The PV address
        plot_style : str, optional
            The style in which to render the data, either 'Line' or 'Bar'
        name : str
            The name of the curve (usually made the same as the PV address)
        color : QColor
            The color for the curve
        lineStyle : str
            The line style of the curve, i.e. solid, dash, dot, etc.
        lineWidth : int
            How thick the curve line should be
        symbol : str
            The symbols as markers along the curve, i.e. circle, square,
            triangle, star, etc.
        symbolSize : int
            How big the symbols should be
        barWidth: float, optional
            Width of any bars drawn on the plot
        upperThreshold: float, optional
            Bars that are above this value will be drawn in the threshold color
        lowerThreshold: float, optional
            Bars that are below this value will be drawn in the threshold color
        thresholdColor: QColor, optional
            Color to draw bars that exceed either threshold
        yAxisName : str
            The name of the y axis to associate with this curve. Will be created if it
            doesn't yet exist

        Returns
        -------
        new_curve : TimePlotCurveItem
            The newly created curve.
        """
        plot_opts = dict()
        plot_opts['symbol'] = symbol
        if symbolSize is not None:
            plot_opts['symbolSize'] = symbolSize
        if lineStyle is not None:
            plot_opts['lineStyle'] = lineStyle
        if lineWidth is not None:
            plot_opts['lineWidth'] = lineWidth

        # Add curve
        new_curve = self.createCurveItem(y_channel=y_channel, plot_by_timestamps=self._plot_by_timestamps,
                                         plot_style=plot_style, name=name, color=color, yAxisName=yAxisName,
                                         useArchiveData=useArchiveData, **plot_opts)
        new_curve.setUpdatesAsynchronously(self.updateMode)
        new_curve.setBufferSize(self._bufferSize)

        self.update_timer.timeout.connect(new_curve.asyncUpdate)
        if plot_style == 'Bar':
            if barWidth is None:
                barWidth = 1.0  # Can't use default since it can be explicitly set to None and avoided
            new_curve.bar_graph_item = BarGraphItem(x=[], height=[], width=barWidth, brush=color)
            new_curve.setBarGraphInfo(barWidth, upperThreshold, lowerThreshold, thresholdColor)
        self.addCurve(new_curve, curve_color=color, y_axis_name=yAxisName)
        if new_curve.bar_graph_item is not None:
            # Must happen after addCurve() so that the view box has been created
            new_curve.getViewBox().addItem(new_curve.bar_graph_item)

        new_curve.data_changed.connect(self.set_needs_redraw)
        self.redraw_timer.start()

        return new_curve

    def createCurveItem(self, y_channel, plot_by_timestamps, plot_style, name,
                        color, yAxisName, useArchiveData, **plot_opts):
        return TimePlotCurveItem(y_channel, plot_by_timestamps=plot_by_timestamps, plot_style=plot_style,
                                 name=name, color=color, yAxisName=yAxisName, **plot_opts)

    def removeYChannel(self, curve):
        """
        Remove a curve from the graph. This also stops update the timer
        associated with the curve.

        Parameters
        ----------
        curve : TimePlotCurveItem
            The curve to be removed.
        """
        self.update_timer.timeout.disconnect(curve.asyncUpdate)
        self.removeCurve(curve)
        if len(self._curves) < 1:
            self.redraw_timer.stop()

    def removeYChannelAtIndex(self, index):
        """
        Remove a curve from the graph, given its index in the graph's curve
        list.

        Parameters
        ----------
        index : int
            The curve's index from the graph's curve list.
        """
        curve = self._curves[index]
        self.removeYChannel(curve)

    @Slot()
    def set_needs_redraw(self):
        self._needs_redraw = True

    @Slot()
    def redrawPlot(self):
        """
        Redraw the graph
        """
        if not self._needs_redraw:
            return

        self.updateXAxis()
        # The minimum and maximum x-axis timestamps visible to the user
        min_x = self.plotItem.getViewBox().state['viewRange'][0][0]
        max_x = self.plotItem.getViewBox().state['viewRange'][0][1]
        for curve in self._curves:
            curve.redrawCurve(min_x=min_x, max_x=max_x)
            self.plot_redrawn_signal.emit(curve)
        self._needs_redraw = False

    def updateXAxis(self, update_immediately=False):
        """
        Update the x-axis for every graph redraw.

        Parameters
        ----------
        update_immediately : bool
            Update the axis range(s) immediately if True, or defer until the
            next rendering.
        """
        if len(self._curves) == 0:
            return

        if self._plot_by_timestamps:
            if self._update_mode == PyDMTimePlot.OnValueChange:
                maxrange = max([curve.max_x() for curve in self._curves])
            else:
                maxrange = time.time()
            minrange = maxrange - self._time_span
            current_min_x = self.plotItem.getAxis('bottom').range[0]  # Minimum x value currently displayed on the plot
            if not self.plotItem.isAnyXAutoRange() or (self.plotItem.isAnyXAutoRange() and
                                                       maxrange - current_min_x >= self._time_span):
                # Keep the rolling window of data moving, unless the user asked for autorange and we've
                # not yet hit the maximum amount of data to display based on the time span
                self.plotItem.setXRange(minrange, maxrange, padding=0.0, update=update_immediately)
        else:
            diff_time = self.starting_epoch_time - max([curve.max_x() for curve in self._curves])
            if diff_time > DEFAULT_X_MIN:
                diff_time = DEFAULT_X_MIN
            self.getViewBox().setLimits(minXRange=diff_time)

    def clearCurves(self):
        """
        Remove all curves from the graph.
        """
        super(PyDMTimePlot, self).clear()

    def getCurves(self):
        """
        Dump the current list of curves and each curve's settings into a list
        of JSON-formatted strings.

        Returns
        -------
        settings : list
            A list of JSON-formatted strings, each containing a curve's
            settings
        """
        return [json.dumps(curve.to_dict()) for curve in self._curves]

    def setCurves(self, new_list):
        """
        Add a list of curves into the graph.

        Parameters
        ----------
        new_list : list
            A list of JSON-formatted strings, each contains a curve and its
            settings
        """
        try:
            new_list = [json.loads(str(i)) for i in new_list]
        except ValueError as e:
            logger.exception("Error parsing curve json data: {}".format(e))
            return
        self.clearCurves()
        for d in new_list:
            color = d.get('color')
            thresholdColor = d.get('thresholdColor')
            if color:
                color = QColor(color)
            if thresholdColor:
                thresholdColor = QColor(thresholdColor)
            self.addYChannel(d['channel'],
                             plot_style=d.get('plot_style'),
                             name=d.get('name'),
                             color=color,
                             lineStyle=d.get('lineStyle'),
                             lineWidth=d.get('lineWidth'),
                             symbol=d.get('symbol'),
                             symbolSize=d.get('symbolSize'),
                             barWidth=d.get('barWidth'),
                             upperThreshold=d.get('upperThreshold'),
                             lowerThreshold=d.get('lowerThreshold'),
                             thresholdColor=thresholdColor,
                             yAxisName=d.get('yAxisName'))

    curves = Property("QStringList", getCurves, setCurves, designable=False)

    def findCurve(self, pv_name):
        """
        Find a curve from a graph's curve list.

        Parameters
        ----------
        pv_name : str
            The curve's PV address.

        Returns
        -------
        curve : TimePlotCurveItem
            The found curve, or None.
        """
        for curve in self._curves:
            if curve.address == pv_name:
                return curve

    def refreshCurve(self, curve):
        """
        Remove a curve currently being plotted on the timeplot, then redraw
        that curve, which could have been updated with a new symbol, line
        style, line width, etc.

        Parameters
        ----------
        curve : TimePlotCurveItem
            The curve to be re-added.
        """
        curve = self.findCurve(curve.channel)
        if curve:
            self.removeYChannel(curve)
            self.addYChannel(y_channel=curve.address, plot_style=curve.plot_style, color=curve.color,
                             name=curve.address, lineStyle=curve.lineStyle,
                             lineWidth=curve.lineWidth, symbol=curve.symbol,
                             symbolSize=curve.symbolSize, barWidth=curve.bar_width,
                             upperThreshold=curve.upper_threshold, lowerThreshold=curve.lower_threshold,
                             thresholdColor=curve.threshold_color, yAxisName=curve.y_axis_name)

    def addLegendItem(self, item, pv_name, force_show_legend=False):
        """
        Add an item into the graph's legend.

        Parameters
        ----------
        item : TimePlotCurveItem
            A curve being plotted in the graph
        pv_name : str
            The PV channel
        force_show_legend : bool
            True to make the legend to be displayed; False to just add the
            item, but do not display the legend.
        """
        self._legend.addItem(item, pv_name)
        self.setShowLegend(force_show_legend)

    def removeLegendItem(self, pv_name):
        """
        Remove an item from the legend.

        Parameters
        ----------
        pv_name : str
            The PV channel, used to search for the legend item to remove.
        """
        self._legend.removeItem(pv_name)
        if len(self._legend.items) == 0:
            self.setShowLegend(False)

    def getBufferSize(self):
        """
        Get the size of the data buffer for the entire chart.

        Returns
        -------
        size : int
            The chart's data buffer size.
        """
        return int(self._bufferSize)

    def setBufferSize(self, value):
        """
        Set the size of the data buffer of the entire chart. This will also
        update the same value for each of the data buffer of each chart's
        curve.

        Parameters
        ----------
        value : int
            The new buffer size for the chart.
        """
        if self._bufferSize != int(value):
            # Originally, the bufferSize is the max between the user's input and 1, and 1 doesn't make sense.
            # So, I'm comparing the user's input with the minimum buffer size, and pick the max between the two
            self._bufferSize = max(int(value), MINIMUM_BUFFER_SIZE)
            for curve in self._curves:
                curve.setBufferSize(value)

    def resetBufferSize(self):
        """
        Reset the data buffer size of the chart, and each of the chart's
        curve's data buffer, to the minimum
        """
        if self._bufferSize != DEFAULT_BUFFER_SIZE:
            self._bufferSize = DEFAULT_BUFFER_SIZE
            for curve in self._curves:
                curve.resetBufferSize()

    bufferSize = Property("int", getBufferSize, setBufferSize, resetBufferSize)

    def getUpdatesAsynchronously(self):
        return self._update_mode == PyDMTimePlot.AtFixedRated

    def setUpdatesAsynchronously(self, value): 
        for curve in self._curves:
            curve.setUpdatesAsynchronously(value)
        """
        Check if value is from updatesAsynchronously(bool) or updateMode(int)
        """
        if type(value)==int and value == updateMode.AtFixedRate or type(value)==bool and value is True: 
            self._update_mode = PyDMTimePlot.AtFixedRate
            self.update_timer.start()
        else:
            self._update_mode = PyDMTimePlot.OnValueChange
            self.update_timer.stop()

    def resetUpdatesAsynchronously(self):
        self._update_mode = PyDMTimePlot.OnValueChange
        self.update_timer.stop()
        for curve in self._curves:
            curve.resetUpdatesAsynchronously()

    updatesAsynchronously = Property("bool",
                                     getUpdatesAsynchronously,
                                     setUpdatesAsynchronously,
                                     resetUpdatesAsynchronously, designable=False)

    @Property(updateMode)
    def updateMode(self):
        """
        The updateMode to be used as property to set plot update mode.

        Returns
        -------
        updateMode
        """
        return self._updateMode

    @updateMode.setter
    def updateMode(self, new_type):
        """
        The updateMode to be used as property to set plot update mode.

        Parameters
        ----------
        new_type : updateMode
        """
        if new_type != self._updateMode:
            self._updateMode = new_type
            self.setUpdatesAsynchronously(self._updateMode)

    def getTimeSpan(self):
        """
        The extent of the x-axis of the chart, in seconds.  In other words,
        how long a data point stays on the plot before falling off the left
        edge.

        Returns
        -------
        time_span : float
            The extent of the x-axis of the chart, in seconds.
        """
        return float(self._time_span)

    def setTimeSpan(self, value):
        """
        Set the extent of the x-axis of the chart, in seconds.
        In aynchronous mode, the chart will allocate enough buffer for the new time span duration.
        Data arriving after each duration will be recorded into the buffer
        having been rotated.

        Parameters
        ----------
        value : float
            The time span duration, in seconds, to allocate enough buffer to
            collect data for, before rotating the buffer.
        """
        value = float(value)
        if self._time_span != value:
            self._time_span = value

            if self.getUpdatesAsynchronously():
                self.setBufferSize(int((self._time_span * 1000.0) / self._update_interval))

            self.updateXAxis(update_immediately=True)

    def resetTimeSpan(self):
        """
        Reset the timespan to the default value.
        """
        if self._time_span != DEFAULT_TIME_SPAN:
            self._time_span = DEFAULT_TIME_SPAN
            if self.getUpdatesAsynchronously():
                self.setBufferSize(int((self._time_span * 1000.0) / self._update_interval))
            self.updateXAxis(update_immediately=True)

    timeSpan = Property(float, getTimeSpan, setTimeSpan, resetTimeSpan)

    def getUpdateInterval(self):
        """
        Get the update interval for the chart.

        Returns
        -------
        interval : float
            The update interval of the chart.
        """
        return float(self._update_interval) / 1000.0

    def setUpdateInterval(self, value):
        """
        Set a new update interval for the chart and update its data buffer size.

        Parameters
        ----------
        value : float
            The new update interval value.
        """
        value = abs(int(1000.0 * value))
        if self._update_interval != value:
            self._update_interval = value
            self.update_timer.setInterval(self._update_interval)
            if self.getUpdatesAsynchronously():
                self.setBufferSize(int((self._time_span * 1000.0) / self._update_interval))

    def resetUpdateInterval(self):
        """
        Reset the chart's update interval to the default.
        """
        if self._update_interval != DEFAULT_UPDATE_INTERVAL:
            self._update_interval = DEFAULT_UPDATE_INTERVAL
            self.update_timer.setInterval(self._update_interval)
            if self.getUpdatesAsynchronously():
                self.setBufferSize(int((self._time_span * 1000.0) / self._update_interval))

    updateInterval = Property(float, getUpdateInterval,
                              setUpdateInterval, resetUpdateInterval)

    def getAutoRangeX(self):
        if self._plot_by_timestamps:
            return False
        return super(PyDMTimePlot, self).getAutoRangeX()

    def setAutoRangeX(self, value):
        if self._plot_by_timestamps:
            self._auto_range_x = False
            self.plotItem.enableAutoRange(ViewBox.XAxis, enable=self._auto_range_x)
        else:
            super(PyDMTimePlot, self).setAutoRangeX(value)

    def channels(self):
        return [curve.channel for curve in self._curves]

    # The methods for autoRangeY, minYRange, and maxYRange are
    # all defined in BasePlot, but we don't expose them as properties there, because not all plot
    # subclasses necessarily want them to be user-configurable in Designer.
    autoRangeY = Property(bool, BasePlot.getAutoRangeY,
                          BasePlot.setAutoRangeY,
                          BasePlot.resetAutoRangeY, doc="""
    Whether or not the Y-axis automatically rescales to fit the data.
    If true, the values in minYRange and maxYRange are ignored.
    """)

    minYRange = Property(float, BasePlot.getMinYRange,
                         BasePlot.setMinYRange, doc="""
    Minimum Y-axis value visible on the plot.""")

    maxYRange = Property(float, BasePlot.getMaxYRange,
                         BasePlot.setMaxYRange, doc="""
    Maximum Y-axis value visible on the plot.""")

    def enableCrosshair(self, is_enabled, starting_x_pos=DEFAULT_X_MIN, starting_y_pos=DEFAULT_Y_MIN, vertical_angle=90,
                        horizontal_angle=0, vertical_movable=False, horizontal_movable=False):
        """
        Display a crosshair on the graph.

        Parameters
        ----------
        is_enabled : bool
            True is to display the crosshair; False is to hide it.
        starting_x_pos : float
            The x position where the vertical line will cross
        starting_y_pos : float
            The y position where the horizontal line will cross
        vertical_angle : int
            The angle of the vertical line
        horizontal_angle : int
            The angle of the horizontal line
        vertical_movable : bool
            True if the user can move the vertical line; False if not
        horizontal_movable : bool
            True if the user can move the horizontal line; False if not
        """
        super(PyDMTimePlot, self).enableCrosshair(is_enabled, starting_x_pos, starting_y_pos, vertical_angle,
                                                  horizontal_angle, vertical_movable, horizontal_movable)


class TimeAxisItem(AxisItem):
    """
    TimeAxisItem formats a unix time axis into a human-readable format.
    """
    def __init__(self, *args, **kwargs):
        super(TimeAxisItem, self).__init__(*args, **kwargs)
        self.enableAutoSIPrefix(False)

    def tickStrings(self, values, scale, spacing):
        strings = []
        for val in values:
            strings.append(time.strftime("%H:%M:%S", time.localtime(val)))
        return strings



from qtpy.QtCore import QAbstractTableModel, Qt, QVariant
from qtpy.QtGui import QBrush
from .baseplot import BasePlotCurveItem


class BasePlotCurvesModel(QAbstractTableModel):
    name_for_symbol = {v: k for k, v in BasePlotCurveItem.symbols.items()}
    name_for_line = {v: k for k, v in BasePlotCurveItem.lines.items()}
    """ This is the data model used by the waveform plot curve editor.
    It basically acts as a go-between for the curves in a plot, and
    QTableView items.
    """

    def __init__(self, plot, parent=None):
        super(BasePlotCurvesModel, self).__init__(parent=parent)
        self._plot = plot
        self._column_names = ("Label", "Color", "Y-Axis Name", "Line Style",
                              "Line Width", "Symbol", "Symbol Size", "Bar Width",
                              "Upper Limit", "Lower Limit", "Limit Color")

    @property
    def plot(self):
        return self._plot

    @plot.setter
    def plot(self, new_plot):
        self.beginResetModel()
        self._plot = new_plot
        self.endResetModel()

    # QAbstractItemModel Implementation
    def clear(self):
        self.plot.clearCurves()

    def flags(self, index):
        column_name = self._column_names[index.column()]
        if column_name == "Color" or column_name == "Limit Color":
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable

    def rowCount(self, parent=None):
        if parent is not None and parent.isValid():
            return 0
        return len(self.plot._curves)

    def columnCount(self, parent=None):
        return len(self._column_names)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        if index.row() >= self.rowCount():
            return QVariant()
        if index.column() >= self.columnCount():
            return QVariant()
        column_name = self._column_names[index.column()]
        curve = self.plot._curves[index.row()]
        if role == Qt.DisplayRole or role == Qt.EditRole:
            return self.get_data(column_name, curve)
        elif role == Qt.BackgroundRole and column_name == "Color":
            return QBrush(curve.color)
        elif role == Qt.BackgroundRole and column_name == "Limit Color":
            return QBrush(curve.threshold_color)
        else:
            return QVariant()

    def get_data(self, column_name, curve):
        if column_name == "Label":
            if curve.name() is None:
                return QVariant()
            return str(curve.name())
        elif column_name == "Y-Axis Name":
            return curve.y_axis_name
        elif column_name == "Color":
            return curve.color_string
        elif column_name == "Limit Color":
            return curve.threshold_color_string
        elif column_name == "Bar Width":
            return curve.bar_width
        elif column_name == "Upper Limit":
            return curve.upper_threshold
        elif column_name == "Lower Limit":
            return curve.lower_threshold
        elif column_name == "Line Style":
            return self.name_for_line[curve.lineStyle]
        elif column_name == "Line Width":
            return int(curve.lineWidth)
        elif column_name == "Symbol":
            return self.name_for_symbol[curve.symbol]
        elif column_name == "Symbol Size":
            return int(curve.symbolSize)

    def setData(self, index, value, role=Qt.EditRole):
        if not index.isValid():
            return False
        if index.row() >= self.rowCount():
            return False
        if index.column() >= self.columnCount():
            return False
        column_name = self._column_names[index.column()]
        curve = self.plot._curves[index.row()]
        if role == Qt.EditRole:
            if isinstance(value, QVariant):
                value = value.toString()
            if not self.set_data(column_name, curve, value):
                return False
        else:
            return False
        self.dataChanged.emit(index, index)
        return True

    def set_data(self, column_name, curve, value):
        if column_name == "Label":
            curve.setData(name=str(value))
        elif column_name == "Color":
            curve.color = value
        elif column_name == "Y-Axis Name":
            curve.y_axis_name = str(value)
        elif column_name == "Line Style":
            curve.lineStyle = int(value)
        elif column_name == "Line Width":
            curve.lineWidth = int(value)
        elif column_name == "Symbol":
            if value is None:
                curve.symbol = None
            else:
                curve.symbol = str(value)
        elif column_name == "Symbol Size":
            curve.symbolSize = int(value)
        elif column_name == "Bar Width":
            curve.bar_width = float(value)
        elif column_name == "Upper Limit":
            curve.upper_threshold = float(value)
        elif column_name == "Lower Limit":
            curve.lower_threshold = float(value)
        elif column_name == "Limit Color":
            curve.threshold_color = value
        else:
            return False
        return True

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role != Qt.DisplayRole:
            return super(BasePlotCurvesModel, self).headerData(
                section, orientation, role)
        if orientation == Qt.Horizontal and section < self.columnCount():
            return str(self._column_names[section])
        elif orientation == Qt.Vertical and section < self.rowCount():
            return section
    # End QAbstractItemModel implementation.

    def append(self, name=None, color=None):
        pass

    def removeAtIndex(self, index):
        pass

    def getColumnIndex(self, column_name):
        """ Returns the column index of the name. Raises a ValueError if it's not a valid column name """
        return self._column_names.index(column_name)

    def needsColorDialog(self, index):
        column_name = self._column_names[index.column()]
        if column_name == "Color" or column_name == "Limit Color":
            return True
        return False



import logging
import os

logger = logging.getLogger(__name__)

# This file gives us flexibility in allowing multiple options for pvAccess plugins in the future if needed.
PVA_LIB = os.getenv("PYDM_PVA_LIB", "").upper()
try:
    if PVA_LIB == "P4P" or not PVA_LIB:
        from pydm.data_plugins.epics_plugins.p4p_plugin_component import P4PPlugin
        PVAPlugin = P4PPlugin
    elif PVA_LIB == "PVAPY":
        logger.error("PVAPY is not currently supported by PyDM")
        PVAPlugin = None
except ImportError:
    PVAPlugin = None
    logger.info("No PVAccess Python library available. Ignoring pva plugin.")

if PVAPlugin:
    PVAPlugin.protocol = "pva"



import os

import entrypoints

from pydm import config
from pydm.data_plugins import (PyDMPlugin, initialize_plugins_if_needed,
                               load_plugins_from_entrypoints,
                               load_plugins_from_path, plugin_for_address,
                               plugin_modules)


def test_data_plugin_add(qapp, test_plugin):
    # Check that adding this after import will be reflected in PyDMApp
    initialize_plugins_if_needed()
    assert isinstance(plugin_modules['tst'], test_plugin)
    assert isinstance(qapp.plugins['tst'], test_plugin)


def test_plugin_directory_loading(qapp, caplog):
    # Create a fake file
    cur_dir = os.getcwd()
    with open(os.path.join(cur_dir, 'plugin_foo.py'), 'w+') as handle:
        handle.write(fake_file)
        handle.flush()

    try:
        # Load plugins. One of them will raise an exception during initialization. This will
        # prevent it from being added to the available plugins, but the others should still load.
        load_plugins_from_path([cur_dir], 'foo.py')
        assert 'tst1' in plugin_modules
        assert 'tst2' in plugin_modules
        assert 'fail' not in plugin_modules

        # Check for the error logged when a plugin fails to load. Confirms that the plugin that failed to load
        # was the one expected to fail, and that there was only one failure total.
        assert "FailingTestPlugin'> failed to load and will not be available for use" in caplog.text
        assert caplog.text.count("failed to load and will not be available for use") == 1
    finally:
        os.remove(os.path.join(cur_dir, 'plugin_foo.py'))


def test_plugin_for_address(test_plugin, monkeypatch):
    # Get by protocol
    assert isinstance(plugin_for_address('tst://tst:this'),
                      test_plugin)
    assert plugin_for_address('tst:this') is None
    # Default protocol
    monkeypatch.setattr(config, 'DEFAULT_PROTOCOL', 'tst')
    assert isinstance(plugin_for_address('tst:this'),
                      test_plugin)


fake_file = """\
from pydm.data_plugins import PyDMPlugin


class TestPlugin1(PyDMPlugin):
    protocol = 'tst1'


class FailingTestPlugin(PyDMPlugin):
    protocol = 'fail'

    def __init__(self, *args, **kwargs):
        raise Exception  # Purposefully fail to initialize for testing purposes

class TestPlugin2(PyDMPlugin):
    protocol = 'tst2'
"""


def test_entrypoint_import(monkeypatch):
    class MyTestPlugin(PyDMPlugin):
        protocol = "__test_suite_protocol__"

    class Entrypoint:
        name = "MyTestPlugin"

        def load(self):
            return MyTestPlugin

    def get_group_all(key):
        yield Entrypoint()

    monkeypatch.setattr(entrypoints, "get_group_all", get_group_all)
    loaded = load_plugins_from_entrypoints()

    assert "__test_suite_protocol__" in loaded
    assert isinstance(loaded["__test_suite_protocol__"], MyTestPlugin)



from qtpy.QtWidgets import (QDialog, QVBoxLayout, QHBoxLayout, QTableView,
                            QAbstractItemView, QSpacerItem, QSizePolicy,
                            QDialogButtonBox, QPushButton, QStyleOptionViewItem, QTabWidget, QWidget,
                            QComboBox, QStyledItemDelegate, QColorDialog, QHeaderView)
from qtpy.QtCore import Qt, Slot, QAbstractItemModel, QModelIndex, QObject, QItemSelection
from qtpy.QtDesigner import QDesignerFormWindowInterface
from .baseplot import BasePlotAxisItem, BasePlotCurveItem
from .baseplot_table_model import BasePlotCurvesModel
from .axis_table_model import BasePlotAxesModel
from collections import OrderedDict
from typing import Optional


class BasePlotCurveEditorDialog(QDialog):
    """QDialog that is used in Qt Designer to edit the properties of the
    curves in a waveform plot.  This dialog is shown when you double-click
    the plot, or when you right click it and choose 'edit curves'.

    This thing is mostly just a wrapper for a table view, with a couple
    buttons to add and remove curves, and a button to save the changes."""
    TABLE_MODEL_CLASS = BasePlotCurvesModel
    AXIS_MODEL_CLASS = BasePlotAxesModel
    AXIS_MODEL_TAB_INDEX = 1

    def __init__(self, plot, parent=None):
        super(BasePlotCurveEditorDialog, self).__init__(parent)
        self.tab_widget = QTabWidget()
        self.plot = plot
        self.setup_ui()
        self.table_model = self.TABLE_MODEL_CLASS(self.plot)
        self.table_view.setModel(self.table_model)
        self.table_model.plot = plot
        self.axis_model = self.AXIS_MODEL_CLASS(self.plot)
        self.axis_view.setModel(self.axis_model)
        self.axis_model.plot = plot
        self.setup_delegate_columns()
        # self.table_view.resizeColumnsToContents()
        self.add_button.clicked.connect(self.addCurve)
        self.remove_button.clicked.connect(self.removeSelectedCurve)
        self.remove_button.setEnabled(False)
        self.add_axis_button.clicked.connect(self.addAxis)
        self.remove_axis_button.clicked.connect(self.removeSelectedAxis)
        self.remove_axis_button.setEnabled(False)
        self.add_axis_count = 0
        self.table_view.selectionModel().selectionChanged.connect(
            self.handleSelectionChange)
        self.axis_view.selectionModel().selectionChanged.connect(
            self.handleSelectionChange)
        self.table_view.doubleClicked.connect(self.handleDoubleClick)
        self.resize(800, 300)

    def setup_ui(self):
        self.vertical_layout = QVBoxLayout(self)
        self.table_view = QTableView(self)
        self.table_view.setEditTriggers(QAbstractItemView.DoubleClicked)
        self.table_view.setProperty("showDropIndicator", False)
        self.table_view.setDragDropOverwriteMode(False)
        self.table_view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_view.setSortingEnabled(False)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.verticalHeader().setVisible(False)
        self.table_view.setColumnWidth(0, 160)
        self.table_view.setColumnWidth(1, 160)
        self.table_view.setColumnWidth(2, 160)
        self.vertical_layout.addWidget(self.tab_widget)
        self.tab_widget.addTab(self.table_view, "Curves")

        self.axis_view = QTableView(self)
        self.axis_view.setEditTriggers(QAbstractItemView.DoubleClicked)
        self.axis_view.setProperty("showDropIndicator", False)
        self.axis_view.setDragDropOverwriteMode(False)
        self.axis_view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.axis_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.axis_view.setSortingEnabled(False)
        self.axis_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.axis_view.verticalHeader().setVisible(False)
        self.tab_widget.addTab(self.axis_view, "Axes")

        self.tab_widget.currentChanged.connect(self.fillAxisData)

        self.add_remove_layout = QHBoxLayout()
        spacer = QSpacerItem(40, 20, QSizePolicy.Expanding,
                             QSizePolicy.Minimum)
        self.add_remove_layout.addItem(spacer)
        self.add_button = QPushButton("Add Curve", self)
        self.add_remove_layout.addWidget(self.add_button)
        self.remove_button = QPushButton("Remove Curve", self)
        self.add_remove_layout.addWidget(self.remove_button)
        self.add_axis_button = QPushButton("Add Axis", self)
        self.add_remove_layout.addWidget(self.add_axis_button)
        self.remove_axis_button = QPushButton("Remove Axis", self)
        self.add_remove_layout.addWidget(self.remove_axis_button)

        # These buttons start out hidden as they are only useful on the axis tab, will be
        # displayed when the user clicks on that tab
        self.add_axis_button.hide()
        self.remove_axis_button.hide()
        self.vertical_layout.addLayout(self.add_remove_layout)
        self.button_box = QDialogButtonBox(self)
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.addButton("Done", QDialogButtonBox.AcceptRole)
        self.vertical_layout.addWidget(self.button_box)
        self.button_box.accepted.connect(self.saveChanges)
        self.button_box.rejected.connect(self.reject)
        self.setWindowTitle("Waveform Curve Editor")

    def setup_delegate_columns(self):
        symbol_delegate = SymbolColumnDelegate(self)
        self.table_view.setItemDelegateForColumn(self.table_model.getColumnIndex('Symbol'), symbol_delegate)
        line_delegate = LineColumnDelegate(self)
        self.table_view.setItemDelegateForColumn(self.table_model.getColumnIndex('Line Style'), line_delegate)
        color_delegate = ColorColumnDelegate(self)
        self.table_view.setItemDelegateForColumn(self.table_model.getColumnIndex('Color'), color_delegate)
        axis_delegate = AxisColumnDelegate(self)
        self.axis_view.setItemDelegateForColumn(self.axis_model.getColumnIndex('Y-Axis Orientation'), axis_delegate)

    @Slot()
    def addCurve(self):
        self.table_model.append()

    @Slot()
    def removeSelectedCurve(self):
        self.table_model.removeAtIndex(self.table_view.currentIndex())

    @Slot()
    def addAxis(self):
        self.add_axis_count += 1
        default_axis_name = 'New Axis ' + str(self.add_axis_count)
        # Just a quick way to ensure that the default named axes are always unique, even when the user closes
        # out a plot widget and re-opens it later
        while default_axis_name in self.plot.plotItem.axes:
            self.add_axis_count += 1
            default_axis_name = 'New Axis ' + str(self.add_axis_count)
        self.axis_model.append(default_axis_name)

    @Slot()
    def removeSelectedAxis(self):
        self.axis_model.removeAtIndex(self.axis_view.currentIndex())

    @Slot(QItemSelection, QItemSelection)
    def handleSelectionChange(self, selected, deselected):
        self.remove_button.setEnabled(
            self.table_view.selectionModel().hasSelection())
        self.remove_axis_button.setEnabled(
            self.axis_view.selectionModel().hasSelection())

    @Slot(QModelIndex)
    def handleDoubleClick(self, index):
        if self.table_model.needsColorDialog(index):
            # The table model returns a QBrush for BackgroundRole, not a QColor
            init_color = self.table_model.data(index,
                                               Qt.BackgroundRole).color()
            color = QColorDialog.getColor(init_color, self)
            if color.isValid():
                self.table_model.setData(index, color, role=Qt.EditRole)

    @Slot()
    def saveChanges(self):
        formWindow = QDesignerFormWindowInterface.findFormWindow(self.plot)
        if formWindow:
            formWindow.cursor().setProperty("yAxes", self.plot.yAxes)
            formWindow.cursor().setProperty("curves", self.plot.curves)
        self.accept()

    @Slot(int)
    def fillAxisData(self, tab_index):
        """ When the user clicks on the axis tab, prefill it with rows based on the curves they have created """

        # Toggle visibility of the buttons every time the tab changes
        self.add_button.setVisible(not self.add_button.isVisible())
        self.remove_button.setVisible(not self.remove_button.isVisible())
        self.add_axis_button.setVisible(not self.add_axis_button.isVisible())
        self.remove_axis_button.setVisible(not self.remove_axis_button.isVisible())

        if tab_index != self.AXIS_MODEL_TAB_INDEX:
            return  # Nothing else to do if this is just the original "curves" tab

        # Fix a display issue on the left axis when editing plots
        if 'left' in self.plot.plotItem.axes:
            self.plot.plotItem.hideAxis('left')

        axis_name_col_index = self.table_model.getColumnIndex('Y-Axis Name')
        curve_axis_names = [str(self.table_model.index(i, axis_name_col_index).data())
                            for i in range(self.table_model.rowCount())]

        existing_axis_names = [str(self.axis_model.index(i, 0).data())
                               for i in range(self.axis_model.rowCount())]

        # Removing duplicates here instead of using a set to preserve order
        names_to_add = []
        for name in curve_axis_names:
            if name not in existing_axis_names and name not in names_to_add:
                names_to_add.append(name)

        for name in names_to_add:
            if name:
                self.axis_model.append(name)


class ColorColumnDelegate(QStyledItemDelegate):
    """The ColorColumnDelegate is an item delegate that is installed on the
    color column of the table view.  Its only job is to ensure that the default
    editor widget (a line edit) isn't displayed for items in the color column.
    """
    def createEditor(self, parent, option, index):
        return None


class AxisColumnDelegate(QStyledItemDelegate):

    """
    AxisColumnDelegate draws a QComboBox with the allowed values for the axis orientation
    column value, which must map to the values expected by PyQtGraph. Helps ensure that the
    user doesn't have to know what these exact values are, and prevents frustrating typos.
    """
    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.addItems(BasePlotAxisItem.axis_orientations.keys())
        return editor

    def setEditorData(self, editor, index):
        val = str(index.model().data(index, Qt.EditRole))
        editor.setCurrentText(val)

    def setModelData(self, editor, model, index):
        val = BasePlotAxisItem.axis_orientations[editor.currentText()]
        model.setData(index, val, Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class LineColumnDelegate(QStyledItemDelegate):
    """LineColumnDelegate draws a QComboBox in the Line Style column, so that users
    can pick the styles they want to display from a list, instead of needing to
    remember the PyQtGraph character codes."""
    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.addItems(BasePlotCurveItem.lines.keys())
        return editor

    def setEditorData(self, editor, index):
        val = str(index.model().data(index, Qt.EditRole))
        editor.setCurrentText(val)

    def setModelData(self, editor, model, index):
        val = BasePlotCurveItem.lines[editor.currentText()]
        model.setData(index, val, Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class SymbolColumnDelegate(QStyledItemDelegate):
    """SymbolColumnDelegate draws a QComboBox in the Symbol column, so that users
    can pick the symbol they want to display from a list, instead of needing to
    remember the PyQtGraph character codes."""
    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.addItems(BasePlotCurveItem.symbols.keys())
        return editor

    def setEditorData(self, editor, index):
        val = str(index.model().data(index, Qt.EditRole))
        editor.setCurrentText(val)

    def setModelData(self, editor, model, index):
        val = BasePlotCurveItem.symbols[editor.currentText()]
        model.setData(index, val, Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class RedrawModeColumnDelegate(QStyledItemDelegate):
    """RedrawModeColumnDelegate draws a QComboBox in the Redraw Mode column, so
    that users can pick the redraw mode from a list."""
    choices = OrderedDict([
        ('X or Y updates', BasePlotCurveItem.REDRAW_ON_EITHER),
        ('Y updates', BasePlotCurveItem.REDRAW_ON_Y),
        ('X updates', BasePlotCurveItem.REDRAW_ON_X),
        ('Both update', BasePlotCurveItem.REDRAW_ON_BOTH)])
    text_for_choices = {v: k for k, v in choices.items()}

    def displayText(self, value, locale):
        return self.text_for_choices[value]

    def createEditor(self, parent, option, index):
        editor = QComboBox(parent)
        editor.addItems(self.choices.keys())
        return editor

    def setEditorData(self, editor, index):
        val = self.text_for_choices[index.model().data(index, Qt.EditRole)]
        editor.setCurrentText(val)

    def setModelData(self, editor, model, index):
        val = self.choices[editor.currentText()]
        model.setData(index, val, Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class PlotStyleColumnDelegate(QStyledItemDelegate):
    """ Allows the user to toggle between line and bar graphs. Hides/shows relevant columns based on that choice. """

    line_columns_to_toggle = ('Line Style', 'Line Width', 'Symbol', 'Symbol Size')
    bar_columns_to_toggle = ('Bar Width', 'Upper Limit', 'Lower Limit', 'Limit Color')

    def __init__(self, parent: QObject, table_model: BasePlotCurvesModel, table_view: QTableView):
        super().__init__(parent)
        self.table_model = table_model
        self.table_view = table_view

    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex) -> QWidget:
        """ Create a combo box that allows the user to choose the style of plot they want. """
        editor = QComboBox(parent)
        editor.addItems(('Line', 'Bar'))
        return editor

    def setEditorData(self, editor: QWidget, index: QModelIndex) -> None:
        val = str(index.model().data(index, Qt.EditRole))
        editor.setCurrentText(val)

    def setModelData(self, editor: QWidget, model: QAbstractItemModel, index: QModelIndex) -> None:
        model.setData(index, editor.currentText(), Qt.EditRole)
        self.toggleColumnVisibility()

    def updateEditorGeometry(self, editor: QWidget, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        editor.setGeometry(option.rect)

    def toggleColumnVisibility(self):
        """ Toggle visibility of columns based on the current state of the associated curve editor table """
        self.hideColumns(hide_line_columns=True, hide_bar_columns=True)
        if len(self.table_model.plot._curves) > 0:
            for curve in self.table_model.plot._curves:
                plot_style = curve.plot_style
                if plot_style is None or plot_style == 'Line':
                    self.hideColumns(hide_line_columns=False)
                elif plot_style == 'Bar':
                    self.hideColumns(hide_bar_columns=False)
        else:
            self.hideColumns(False, True)  # Show line columns only as a default

    def hideColumns(self, hide_line_columns: Optional[bool] = None, hide_bar_columns: Optional[bool] = None) -> None:
        """ Show or hide columns related to a specific plot style based on the input. If an input parameter
            is omitted (or explicitly set to None), the associated columns will be left alone. """

        if hide_line_columns is not None:
            for column in self.line_columns_to_toggle:
                self.table_view.setColumnHidden(self.table_model.getColumnIndex(column), hide_line_columns)
        if hide_bar_columns is not None:
            for column in self.bar_columns_to_toggle:
                self.table_view.setColumnHidden(self.table_model.getColumnIndex(column), hide_bar_columns)



import json
import time
import numpy as np
from collections import OrderedDict
from typing import List, Optional, Union
from pyqtgraph import DateAxisItem, ErrorBarItem, ViewBox
from pydm.widgets.channel import PyDMChannel
from pydm.widgets.timeplot import TimePlotCurveItem
from pydm.widgets import PyDMTimePlot
from qtpy.QtCore import QObject, QTimer, Property, Signal, Slot
from qtpy.QtGui import QColor

import logging
logger = logging.getLogger(__name__)

DEFAULT_ARCHIVE_BUFFER_SIZE = 18000
DEFAULT_TIME_SPAN = 5.0

class ArchivePlotCurveItem(TimePlotCurveItem):
    """
    ArchivePlotCurveItem is a TimePlotCurveItem with support for receiving data from
    the archiver appliance.

    Parameters
    ----------
    channel_address : str
        The address to of the scalar data to plot. Will also be used to retrieve data
        from archiver appliance if requested.
    use_archive_data : bool
        If True, requests will be made to archiver appliance for archived data when
        the plot is zoomed or scrolled to the left.
    **kws : dict[str: any]
        Additional parameters supported by pyqtgraph.PlotDataItem.
    """

    # Used to request data from archiver appliance (starting timestamp, ending timestamp, processing command)
    archive_data_request_signal = Signal(float, float, str)
    archive_data_received_signal = Signal()

    def __init__(self, channel_address: Optional[str] = None, use_archive_data: bool = True, **kws):
        super(ArchivePlotCurveItem, self).__init__(channel_address, **kws)
        self.use_archive_data = use_archive_data
        self.archive_channel = None
        self.archive_points_accumulated = 0
        self._archiveBufferSize = DEFAULT_ARCHIVE_BUFFER_SIZE
        self.archive_data_buffer = np.zeros((2, self._archiveBufferSize), order='f', dtype=float)

        # When optimized or mean value data is requested, we can display error bars representing
        # the full range of values retrieved
        self.error_bar_item = ErrorBarItem()
        self.error_bar_needs_set = True

        if channel_address is not None and use_archive_data:
            self.setArchiveChannel(channel_address)

    def to_dict(self) -> OrderedDict:
        """ Returns an OrderedDict representation with values for all properties needed to recreate this curve. """
        dic_ = OrderedDict([("useArchiveData", self.use_archive_data), ])
        dic_.update(super(ArchivePlotCurveItem, self).to_dict())
        return dic_

    def setArchiveChannel(self, address: str) -> None:
        """ Creates the channel for the input address for communicating with the archiver appliance plugin. """
        archiver_prefix = 'archiver://pv='
        if address.startswith('ca://'):
            archive_address = address.replace('ca://', archiver_prefix, 1)
        elif address.startswith('pva://'):
            archive_address = address.replace('pva://', archiver_prefix, 1)
        else:
            archive_address = archiver_prefix + address

        self.archive_channel = PyDMChannel(address=archive_address,
                                           value_slot=self.receiveArchiveData,
                                           value_signal=self.archive_data_request_signal)

    @Slot(np.ndarray)
    def receiveArchiveData(self, data: np.ndarray) -> None:
        """ Receive data from archiver appliance and place it into the archive data buffer.
            Will overwrite any previously existing data at the indices written to.

            Parameters
            ----------
            data : np.ndarray
                A numpy array of varying shape consisting of archived data for display.
                At a minimum, index 0 will contain the timestamps and index 1 the actual data observations.
                Additional indices may be used as well based on the type of request made to the archiver appliance.
                For example optimized data will include standard deviations, minimums, and maximums
        """
        archive_data_length = len(data[0])
        max_x = data[0][archive_data_length-1]

        if self.points_accumulated != 0:
            while max_x > self.data_buffer[0][-self.points_accumulated]:
                # Sometimes optimized queries return data past the current timestamp, this will delete those data points
                data = np.delete(data, len(data[0]) - 1, axis=1)
                archive_data_length -= 1
                max_x = data[0][archive_data_length-1]

        self.archive_data_buffer[0, len(self.archive_data_buffer[0]) - archive_data_length:] = data[0]
        self.archive_data_buffer[1, len(self.archive_data_buffer[0]) - archive_data_length:] = data[1]
        self.archive_points_accumulated = archive_data_length

        # Error bars
        if data.shape[0] == 5:  # 5 indicates optimized data was requested from the archiver
            self.error_bar_item.setData(x=self.archive_data_buffer[0, -self.archive_points_accumulated:],
                                        y=self.archive_data_buffer[1, -self.archive_points_accumulated:],
                                        top=data[4] - data[1],
                                        bottom=data[1] - data[3],
                                        beam=0.5,
                                        pen={'color': self.color})
            if self.error_bar_needs_set:
                self.getViewBox().addItem(self.error_bar_item)
                self.error_bar_needs_set = False

        self.data_changed.emit()
        self.archive_data_received_signal.emit()

    def insert_archive_data(self, data: np.ndarray) -> None:
        """
        Inserts data directly into the archive buffer.

        An example use case would be zooming into optimized mean-value data and
        replacing it with the raw data.

        Parameters
        ----------
        data : np.ndarray
           A numpy array of shape (2, length_of_data). Index 0 contains
           timestamps and index 1 contains the data observations.
        """
        archive_data_length = len(data[0])
        min_x = data[0][0]
        max_x = data[0][archive_data_length-1]
        # Get the indices between which we want to insert the data
        min_insertion_index = np.searchsorted(self.archive_data_buffer[0], min_x)
        max_insertion_index = np.searchsorted(self.archive_data_buffer[0], max_x)
        # Delete any non-raw data between the indices so we don't have multiple data points for the same timestamp
        self.archive_data_buffer = np.delete(self.archive_data_buffer,
                                             slice(min_insertion_index, max_insertion_index),
                                             axis=1)
        num_points_deleted = max_insertion_index - min_insertion_index
        delta_points = archive_data_length - num_points_deleted
        if archive_data_length > num_points_deleted:
            # If the insertion will overflow the data buffer, need to delete the oldest points
            self.archive_data_buffer = np.delete(self.archive_data_buffer, slice(0, delta_points), axis=1)
        else:
            self.archive_data_buffer = np.insert(self.archive_data_buffer, [0], np.zeros((2, delta_points)), axis=1)
        min_insertion_index = np.searchsorted(self.archive_data_buffer[0], min_x)
        self.archive_data_buffer = np.insert(self.archive_data_buffer, [min_insertion_index], data[0:2], axis=1)

        self.archive_points_accumulated += archive_data_length - num_points_deleted

    @Slot()
    def redrawCurve(self, min_x=None, max_x=None) -> None:
        """
        Redraw the curve with any new data added since the last draw call.
        """
        if self.archive_points_accumulated == 0:
            super(ArchivePlotCurveItem, self).redrawCurve()
        else:
            try:
                x = np.concatenate((self.archive_data_buffer[0, -self.archive_points_accumulated:].astype(float),
                                    self.data_buffer[0, -self.points_accumulated:].astype(float)))

                y = np.concatenate((self.archive_data_buffer[1, -self.archive_points_accumulated:].astype(float),
                                    self.data_buffer[1, -self.points_accumulated:].astype(float)))

                self.setData(y=y, x=x)
            except (ZeroDivisionError, OverflowError, TypeError):
                # Solve an issue with pyqtgraph and initial downsampling
                pass

    def initializeArchiveBuffer(self) -> None:
        """
        Initialize the archive data buffer used for this curve.
        """
        self.archive_points_accumulated = 0
        self.archive_data_buffer = np.zeros((2, self._archiveBufferSize), order='f', dtype=float)

    def getArchiveBufferSize(self) -> int:
        """ Return the length of the archive buffer """
        return int(self._archiveBufferSize)

    def setArchiveBufferSize(self, value: int) -> None:
        """ Set the length of the archive data buffer and zero it out """
        if self._archiveBufferSize != int(value):
            self._archiveBufferSize = max(int(value), 2)
            self.initializeArchiveBuffer()

    def resetArchiveBufferSize(self) -> None:
        """ Reset the length of the archive buffer back to the default and zero it out """
        if self._archiveBufferSize != DEFAULT_ARCHIVE_BUFFER_SIZE:
            self._archiveBufferSize = DEFAULT_ARCHIVE_BUFFER_SIZE
            self.initializeArchiveBuffer()

    def channels(self) -> List[PyDMChannel]:
        """ Return the list of channels this curve is connected to """
        return [self.channel, self.archive_channel]


class PyDMArchiverTimePlot(PyDMTimePlot):
    """
    PyDMArchiverTimePlot is a PyDMTimePlot with support for receiving data from
    the archiver appliance.

    Parameters
    ----------
    parent : QObject, optional
        The parent of this widget.
    init_y_channels : list
        A list of scalar channels to plot vs time.
    background: str
        The background color for the plot.  Accepts any arguments that
        pyqtgraph.mkColor will accept.
    optimized_data_bins: int
        The number of bins of data returned from the archiver when using optimized requests
    """

    def __init__(self, parent: Optional[QObject] = None, init_y_channels: List[str] = [],
                 background: str = 'default', optimized_data_bins: int = 2000):
        super(PyDMArchiverTimePlot, self).__init__(parent=parent, init_y_channels=init_y_channels,
                                                   plot_by_timestamps=True, background=background,
                                                   bottom_axis=DateAxisItem('bottom'))
        self.optimized_data_bins = optimized_data_bins
        self._min_x = None
        self._prev_x = None  # Holds the minimum x-value of the previous update of the plot
        self._starting_timestamp = time.time()  # The timestamp at which the plot was first rendered
        self._archive_request_queued = False

    def updateXAxis(self, update_immediately: bool = False) -> None:
        """ Manages the requests to archiver appliance. When the user pans or zooms the x axis to the left,
            a request will be made for backfill data """
        if len(self._curves) == 0:
            return

        min_x = self.plotItem.getAxis('bottom').range[0]  # Gets the leftmost timestamp displayed on the x-axis
        max_x = max([curve.max_x() for curve in self._curves])
        max_range = self.plotItem.getAxis('bottom').range[1]
        if min_x == 0:  # This is zero when the plot first renders
            min_x = time.time()
            self._min_x = min_x
            self._starting_timestamp = min_x - DEFAULT_TIME_SPAN  # A bit of a buffer so we don't overwrite live data
            if self.getTimeSpan() != DEFAULT_TIME_SPAN:
                # Initialize x-axis based on the time span as well as trigger a call to the archiver below
                self._min_x = self._min_x - self.getTimeSpan()
                self._archive_request_queued = True
                self.requestDataFromArchiver()
            self.plotItem.setXRange(self._min_x, time.time(), padding=0.0, update=update_immediately)
        elif min_x < self._min_x and not self.plotItem.isAnyXAutoRange():
            # This means the user has manually scrolled to the left, so request archived data
            self._min_x = min_x
            self.setTimeSpan(max_x - min_x)
            if not self._archive_request_queued:
                # Letting the user pan or scroll the plot is convenient, but can generate a lot of events in under
                # a second that would trigger a request for data. By using a timer, we avoid this burst of events
                # and consolidate what would be many requests to archiver into just one.
                self._archive_request_queued = True
                QTimer.singleShot(1000, self.requestDataFromArchiver)
        # Here we only update the x-axis if the user hasn't asked for autorange and they haven't zoomed in (as
        # detected by the max range showing on the plot being less than the data available)
        elif not self.plotItem.isAnyXAutoRange() and not max_range < max_x - 10:
            if min_x > (self._prev_x + 15) or min_x < (self._prev_x - 15):
                # The plus/minus 15 just makes sure we don't do this on every update tick of the graph
                self.setTimeSpan(max_x - min_x)
            else:
                # Keep the plot moving with a rolling window based on the current timestamp
                self.plotItem.setXRange(max_x - self.getTimeSpan(), max_x, padding=0.0, update=update_immediately)
        self._prev_x = min_x

    def requestDataFromArchiver(self, min_x: Optional[float] = None, max_x: Optional[float] = None) -> None:
        """
        Make the request to the archiver appliance data plugin for archived data.

        Parameters
        ----------
        min_x : float, optional
           Timestamp representing the start of the time period to fetch archive data from. Defaults
           to the minimum value visible on the plot when omitted.
        max_x : float, optional
           Timestamp representing the end of the time period to fetch archive data from. Defaults
           to the timestamp of the oldest live data point in the buffer if available. If no live points are
           recorded yet, then defaults to the timestamp at which the plot was first rendered.
        """
        processing_command = ''
        if min_x is None:
            min_x = self._min_x
        for curve in self._curves:
            if curve.use_archive_data:
                if max_x is None:
                    if curve.points_accumulated > 0:
                        max_x = curve.data_buffer[0][curve.getBufferSize() - curve.points_accumulated]
                    else:
                        max_x = self._starting_timestamp
                requested_seconds = max_x - min_x
                if requested_seconds <= 5:
                    self._archive_request_queued = False
                    continue  # Avoids noisy requests when first rendering the plot
                # Max amount of raw data to return before using optimized data
                max_data_request = int(0.80 * self.getArchiveBufferSize())
                if requested_seconds > max_data_request:
                    processing_command = 'optimized_' + str(self.optimized_data_bins)
                curve.archive_data_request_signal.emit(min_x, max_x - 1, processing_command)

    def getArchiveBufferSize(self) -> int:
        """ Returns the size of the data buffer used to store archived data """
        if len(self._curves) == 0:
            return DEFAULT_ARCHIVE_BUFFER_SIZE
        return self._curves[0].getArchiveBufferSize()

    def createCurveItem(self, y_channel: str, plot_by_timestamps: bool, name: str, color: Union[QColor, str],
                        yAxisName: str, useArchiveData: bool, **plot_opts) -> ArchivePlotCurveItem:
        """ Create and return a curve item to be plotted """
        curve_item = ArchivePlotCurveItem(y_channel, use_archive_data=useArchiveData, plot_by_timestamps=plot_by_timestamps,
                                          name=name, color=color, yAxisName=yAxisName, **plot_opts)
        curve_item.archive_data_received_signal.connect(self.archive_data_received)
        return curve_item

    @Slot()
    def archive_data_received(self):
        """ Take any action needed when this plot receives new data from archiver appliance """
        max_x = max([curve.max_x() for curve in self._curves])
        # Assure the user sees all data available whenever the request data is returned
        self.plotItem.setXRange(max_x - self.getTimeSpan(), max_x, padding=0.0, update=True)
        self._archive_request_queued = False

    def setTimeSpan(self, value):
        """ Set the value of the plot's timespan """
        if value < DEFAULT_TIME_SPAN:  # Less than 5 seconds will break the plot
            return
        self._time_span = value

    def clearCurves(self) -> None:
        """ Clear all curves from the plot """
        for curve in self._curves:
            # Need to clear out any bars from optimized data, then super() can handle the rest
            if not curve.error_bar_needs_set:
                curve.getViewBox().removeItem(curve.error_bar_item)

        # reset _min_x to let updateXAxis make requests anew
        self._min_x = self._starting_timestamp 
        super().clearCurves()

    def getCurves(self) -> List[str]:
        """
        Dump and return the current list of curves and each curve's settings into a list
        of JSON-formatted strings.
        """
        return [json.dumps(curve.to_dict()) for curve in self._curves]

    def setCurves(self, new_list: List[str]) -> None:
        """
        Add a list of curves into the graph.

        Parameters
        ----------
        new_list : list
            A list of JSON-formatted strings, each contains a curve and its
            settings
        """
        try:
            new_list = [json.loads(str(i)) for i in new_list]
        except ValueError as e:
            logger.exception("Error parsing curve json data: {}".format(e))
            return
        self.clearCurves()
        for d in new_list:
            color = d.get('color')
            if color:
                color = QColor(color)
            self.addYChannel(d['channel'],
                             name=d.get('name'),
                             color=color,
                             lineStyle=d.get('lineStyle'),
                             lineWidth=d.get('lineWidth'),
                             symbol=d.get('symbol'),
                             symbolSize=d.get('symbolSize'),
                             yAxisName=d.get('yAxisName'),
                             useArchiveData=d.get('useArchiveData'))

    curves = Property("QStringList", getCurves, setCurves, designable=False)



import math
import numpy as np
from typing import Any

import logging
import warnings

logger = logging.getLogger(__name__)


class DisplayFormat(object):
    """Display format for showing data in a PyDM widget."""
    #: The default display format.
    Default = 0
    #: Show the data as a string.
    String = 1
    #: Show numerical values as floating point (base 10, decimal) values.
    Decimal = 2
    #: Show numerical values in scientific / exponential notation.
    Exponential = 3
    #: Show numerical values in base 16 (hexadecimal) notation.
    Hex = 4
    #: Show numerical values in base 2 (binary) notation.
    Binary = 5


def parse_value_for_display(
    value: Any,
    precision: int,
    display_format_type: int = DisplayFormat.Default,
    string_encoding: str = "utf_8",
    widget=None,
):
    """
    Format a value to show it in a widget, based on the display format type.

    Parameters
    ----------
    value : Any
        The value to convert to a string.
    precision : int
        Precision of floating point values to use.
    display_format_type : int, optional
        Display format type to use.
    string_encoding : str, optional
        Encoding to use for strings.
    widget : QtWidgets.QWidget, optional
        Widget to get a name from for conversion errors.

    Returns
    -------
    str
        Formatted version of ``value``.
    """
    if value is None:
        return ""
    try:
        widget_name = widget.objectName()
    except(AttributeError, TypeError):
        widget_name = ""

    if display_format_type == DisplayFormat.Default:
        return value
    elif display_format_type == DisplayFormat.String:
        if isinstance(value, np.ndarray):
            try:
                # Stop at the first zero (EPICS convention)
                # Assume the ndarray is one-dimensional
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    zeros = np.where(value == 0)[0]
                if zeros.size > 0:
                    value = value[:zeros[0]]
                r = value.tobytes().decode(string_encoding)
            except Exception:
                logger.error("Could not decode {0} using {1} at widget named '{2}'.".format(
                    value, string_encoding, widget_name))
                return value
            return r
        else:
            return value
    elif display_format_type == DisplayFormat.Decimal:
        # This case is taken care by the current string formatting
        # routine
        return value
    elif display_format_type == DisplayFormat.Exponential:
        fmt_string = "{" + ":.{}e".format(precision) + "}"
        try:
            r = fmt_string.format(value)
        except (ValueError, TypeError):
            logger.error("Could not display value '{0}' using displayFormat 'Exponential' at widget named "
                         "'{1}'.".format(value, widget_name))
            r = value
        return r
    elif display_format_type == DisplayFormat.Hex:
        try:
            r = hex(int(math.floor(value)))
        except (ValueError, TypeError):
            logger.error("Could not display value '{0}' using displayFormat 'Hex' at widget named "
                         "'{1}'.".format(value, widget_name))
            r = value
        return r
    elif display_format_type == DisplayFormat.Binary:
        try:
            r = bin(int(math.floor(value)))
        except (ValueError, TypeError):
            logger.error("Could not display value '{0}' using displayFormat 'Binary' at widget named "
                         "'{1}'.".format(value, widget_name))
            r = value
        return r



import os
import pytest
import sys
from qtpy.QtWidgets import QApplication

test_ui_path_with_relative_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "../test_data", "test_relative_filename_parent.ui")

def test_show_with_relative_filename(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.open(test_ui_path_with_relative_path)
    main_window.setWindowTitle("Embedded Display Test")
    qtbot.addWidget(main_window)
    display = main_window.home_widget.embeddedDisplay
    # Default behavior should be to not follow symlinks (for backwards compat.).
    # Same effect as: display.followSymlinks = False
    def check_embed():
        assert display.embedded_widget is not None
    qtbot.waitUntil(check_embed)

@pytest.mark.skipif(sys.platform == "win32" and sys.version_info < (3, 8), reason="os.path.realpath on Python 3.7 and prior does not resolve symlinks on Windows")
def test_show_with_relative_filename_and_symlink(qtbot, tmp_path):
    symlinked_ui_file = tmp_path / "test_ui_with_relative_path.ui"
    try:
        os.symlink(test_ui_path_with_relative_path, symlinked_ui_file)
    except:
        pytest.skip("Unable to create a symlink for testing purposes.")

    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.open(symlinked_ui_file)
    main_window.setWindowTitle("Embedded Display Test")
    qtbot.addWidget(main_window)
    display = main_window.home_widget.embeddedDisplay
    display.followSymlinks = True
    def check_embed():
        assert display.embedded_widget is not None
    qtbot.waitUntil(check_embed)



import logging
import numpy as np
from operator import itemgetter
from pydm.widgets.base import PyDMWidget, PyDMWritableWidget
from qtpy import QtCore, QtWidgets

logger = logging.getLogger(__name__)


class PythonTableModel(QtCore.QAbstractTableModel):
    def __init__(self, column_names, initial_list=None, parent=None,
                 edit_method=None, can_edit_method=None):
        super().__init__(parent=parent)
        self.parent = parent
        self._list = None
        self._column_names = column_names
        self.edit_method = edit_method
        self.can_edit_method = can_edit_method
        self.list = initial_list

    @property
    def list(self):
        return self._list

    @list.setter
    def list(self, new_list):
        if new_list is None:
            new_list = []
        self.beginResetModel()
        self._list = list(new_list)
        self.endResetModel()

    # QAbstractItemModel Implementation
    def clear(self):
        self.list = []

    def flags(self, index):
        f = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if self.edit_method is not None:
            editable = True
            if self.can_edit_method is not None:
                editable = self.can_edit_method(
                    self._list[index.row()][index.column()])
            if editable:
                f = f | QtCore.Qt.ItemIsEditable
        return f

    def rowCount(self, parent=None):
        if parent is not None and parent.isValid():
            return 0
        return len(self._list)

    def columnCount(self, parent=None):
        return len(self._column_names)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return QtCore.QVariant()
        if index.row() >= self.rowCount():
            return QtCore.QVariant()
        if index.column() >= self.columnCount():
            return QtCore.QVariant()
        if role == QtCore.Qt.DisplayRole:
            try:
                item = str(self._list[index.row()][index.column()])
            except IndexError:
                item = ""
            return item
        else:
            return QtCore.QVariant()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if self.edit_method is None:
            return False
        if role != QtCore.Qt.EditRole:
            return False
        if not index.isValid():
            return False
        if index.row() >= self.rowCount():
            return False
        if index.column() >= self.columnCount():
            return False

        success = self.edit_method(self.parent, index.row(), index.column(), value)
        
        if success:
            self.dataChanged.emit(index, index)
        return success

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role != QtCore.Qt.DisplayRole:
            return super(PythonTableModel, self).headerData(section,
                                                            orientation, role)
        if orientation == QtCore.Qt.Horizontal \
                and section < self.columnCount():
            return str(self._column_names[section])
        elif orientation == QtCore.Qt.Vertical and section < self.rowCount():
            return section

    def sort(self, col, order=QtCore.Qt.AscendingOrder):
        self.layoutAboutToBeChanged.emit()
        sort_reversed = (order == QtCore.Qt.AscendingOrder)
        self._list.sort(key=itemgetter(col), reverse=sort_reversed)
        self.layoutChanged.emit()

    # End QAbstractItemModel implementation.

    # Python collection implementation.
    def __len__(self):
        return len(self._list)

    def __iter__(self):
        return iter(self._list)

    def __contains__(self, value):
        return value in self._list

    def __getitem__(self, index):
        return self._list[index]

    def __setitem__(self, index, value):
        if len(value) != self.columnCount():
            msg = "Items must have the same length as the column count ({})"
            raise ValueError(msg.format(self.columnCount()))
        self._list[index] = value
        self.dataChanged.emit(index, index)

    def __delitem__(self, index):
        if (index + 1) > len(self):
            raise IndexError("list assignment index out of range")
        self.beginRemoveRows(QtCore.QModelIndex(), index, index)
        del self._list[index]
        self.endRemoveRows()

    def append(self, value):
        self.beginInsertRows(QtCore.QModelIndex(), len(self._list),
                             len(self._list))
        self._list.append(value)
        self.endInsertRows()

    def extend(self, values):
        self.beginInsertRows(QtCore.QModelIndex(), len(self._list),
                             len(self._list) + len(values) - 1)
        self._list.extend(values)
        self.endInsertRows()

    def remove(self, item):
        index = None
        try:
            index = self._list.index(item)
        except ValueError:
            raise ValueError("list.remove(x): x not in list")
        del self[index]

    def pop(self, index=None):
        if len(self._list) < 1:
            raise IndexError("pop from empty list")
        if index is None:
            index = len(self._list) - 1
        del self[index]

    def count(self, item):
        return self._list.count(item)

    def reverse(self):
        self.layoutAboutToBeChanged.emit()
        self._list.reverse()
        self.layoutChanged.emit()


class PyDMNTTable(QtWidgets.QWidget, PyDMWritableWidget):
    """
    The PyDMNTTable is a table widget used to display PVA NTTable data. 

    The PyDMNTTable has two ways of filling the table from the data. 
    If the incoming data dictionary has a 'labels' and/or a 'value' key. 
    Then the list of labels will be set with the data from the 'labels' key. 
    While the data from the 'value' key will be used to set the values in the table. 
    if neither 'labels' or 'value' key are present in the incoming 'data' dictionary,
    then the keys of the data dictionary are set as the labels 
    and all the values stored by the keys will make up the values of the table. 
    
    Parameters
        ----------
        parent : QWidget, optional
            The parent widget for the PyDMNTTable
        init_channel : str, optional
            The channel to be used by the widget. 
    """
    def __init__(self, parent=None, init_channel=None):
        self._read_only = True 

        super().__init__(parent=parent)
        PyDMWidget.__init__(self, init_channel=init_channel)
        self.setLayout(QtWidgets.QVBoxLayout())
        self._table = QtWidgets.QTableView(self)
        self.layout().addWidget(self._table)
        self._model = None
        self._table_labels = None
        self._table_values = []
        self.edit_method = None

    @QtCore.Property(bool)
    def readOnly(self):
        return self._read_only

    @readOnly.setter
    def readOnly(self, value):
        if self._read_only != value:
            self._read_only = value
        
    def check_enable_state(self):
        """
        Checks whether or not the widget should be disable.

        """
        PyDMWritableWidget.check_enable_state(self)
        self.setEnabled(True)
        tooltip = self.toolTip()

        if self.readOnly: 
            if tooltip != '':
                tooltip += '\n'
            tooltip += "Running PyDMNTTable on Read-Only mode."
        
        self.setToolTip(tooltip)

    def value_changed(self, data=None):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        data : dict
            The new value from the channel. 
        """
        if data is None:
            return
        
        super(PyDMNTTable, self).value_changed(data)
    
        labels = data.get('labels', None)
        values = data.get('value', {})

        if not values: 
            values = data.values()
        
        if labels is None or len(labels) == 0:
            labels = data.keys()
            labels = list(labels)
        
        try:
            values = list(zip(*[v for k, v in data.items() if k != 'labels']))
        except TypeError:
            logger.exception("NTTable value items must be iterables.")

        self._table_values = values

        if labels != self._table_labels:
            
            if not self.readOnly:
                self.edit_method = PyDMNTTable.send_table
            else:
                self.edit_method = None

            self._table_labels = labels
            self._model = PythonTableModel(labels, 
                                           initial_list=values, 
                                           parent=self,
                                           edit_method=self.edit_method)
            self._table.setModel(self._model)
        else:
            self._model.list = values

    def send_table(self, row, column, value):
        """
        Update Channel value when cell value is changed.
        
        Parameters
        ----------
        row : int 
            index of row
        column : int 
            index of column 
        value : str
            new value of cell
        """
        if isinstance(self.value[self._table_labels[column]], np.ndarray):
            self.value[self._table_labels[column]] = self.value[self._table_labels[column]].copy()

        if isinstance(self.value[self._table_labels[column]][row], np.bool_):
            if value == "True":
                self.value[self._table_labels[column]][row] = True
            elif value == "False":
                self.value[self._table_labels[column]][row] = False
        else:
            self.value[self._table_labels[column]][row] = value

        value_to_send = {k: v for k, v in self.value.items() if k != 'labels'}

        # dictionary needs to be wrapped in another dictionary with a key 'value'
        # to be passed back to the p4p plugin. 
        emit_dict = {'value': value_to_send}
        
        self.send_value_signal[dict].emit(emit_dict)
        return True



from pydm.show_macros import MacroWindow


def test_macro_window_launches(qtbot):
    """Make sure the Macro window doesn't crash."""
    a = MacroWindow(parent=None)
    qtbot.addWidget(a)
    a.show()



import os
import pytest
import logging
import json
import copy
import webbrowser

from qtpy import QtCore
from qtpy.QtWidgets import QMessageBox, QTableWidgetSelectionRange

from ...widgets.symbol import PyDMSymbol
from ...widgets.symbol_editor import SymbolEditor


def test_symbol_editor(qtbot, monkeypatch):
    """
    Test the rules editor in general.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    monkeypatch : fixture
        To override dialog behaviors
    """
    # Create the base widget
    widget = PyDMSymbol()
    qtbot.addWidget(widget)

    # Ensure that no rules are set
    assert len(widget._state_images) == 0

    # Create a rule editor for this widget
    empty = SymbolEditor(widget)
    qtbot.addWidget(empty)

    empty.show()

    # Abort the changes
    empty.cancelChanges()

    # Create the rules data for the widget
    symbol_dict = {"1":"goodbye.jpg"}

    # Add the rules to the widget
    widget.imageFiles = json.dumps(symbol_dict)

    # Create a new Editor Window
    se = SymbolEditor(widget)
    qtbot.addWidget(se)
    se.show()

    assert se.tbl_symbols.rowCount() == 1
    assert not se.frm_edit.isEnabled()

    se.tbl_symbols.setCurrentCell(0, 0)
    assert se.frm_edit.isEnabled()
    assert se.txt_state.text() == "1"
    assert se.txt_file.text() == "goodbye.jpg"

    qtbot.mouseClick(se.btn_add_symbol, QtCore.Qt.LeftButton)
    assert se.txt_state.text() == "New State"
    assert se.txt_file.text() == "New File"
    assert se.lbl_image.text() == "Could not load image \nNew File"
    assert se.frm_edit.isEnabled()

    qtbot.keyClicks(se.txt_state, "-Test")
    qtbot.keyClick(se.txt_state, QtCore.Qt.Key_Return)
    assert se.txt_state.text() == "New State-Test"
    assert "New State-Test" in se.symbols
    assert se.symbols["New State-Test"] == "New File"

    qtbot.keyClick(se.txt_file, "!")
    assert se.txt_file.text() == "New File!"
    assert se.symbols["New State-Test"] == "New File!"

    # Test Delete Symbol with Confirm - NO
    assert se.tbl_symbols.rowCount() == 2
    se.tbl_symbols.setRangeSelected(QTableWidgetSelectionRange(1, 0, 1, 1), True)
    monkeypatch.setattr(QMessageBox, 'question', lambda *args: QMessageBox.No)
    qtbot.mouseClick(se.btn_del_symbol, QtCore.Qt.LeftButton)
    assert se.tbl_symbols.rowCount() == 2
    assert se.frm_edit.isEnabled()

    # Test Delete Symbol with Confirm - YES
    se.tbl_symbols.setRangeSelected(QTableWidgetSelectionRange(1, 0, 1, 1), True)
    monkeypatch.setattr(QMessageBox, 'question', lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(se.btn_del_symbol, QtCore.Qt.LeftButton)
    assert se.tbl_symbols.rowCount() == 1
    assert len(se.symbols) == 1
    assert se.frm_edit.isEnabled() is False

    # Test Delete Symbol with No Selection
    se.tbl_symbols.clearSelection()
    monkeypatch.setattr(QMessageBox, 'question', lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(se.btn_del_symbol, QtCore.Qt.LeftButton)
    assert se.tbl_symbols.rowCount() == 1
    assert len(se.symbols) == 1
    assert se.frm_edit.isEnabled() is False

def test_symbol_editor_data_valid(qtbot):
    """
    Test the rules form validation.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    """
    def validate(expected_status, expected_msg):
        status, msg = se.is_data_valid()
        assert status == expected_status
        assert expected_msg in msg

    # Create the base widget
    widget = PyDMSymbol()
    qtbot.addWidget(widget)

    # Create the rules data for the widget
    symbol_dict = {"1":"goodbye.jpg"}

    # Add the rules to the widget
    widget.imageFiles = json.dumps(symbol_dict)

    # Ensure that no rules are set
    assert widget.imageFiles is not None

    # Create a rule editor for this widget
    se = SymbolEditor(widget)
    qtbot.addWidget(se)

    # can use this when have images set up
    # validate(True, '')

    symbols_original = copy.deepcopy(se.symbols)

    se.symbols[""] = "goodbye.jpg"
    validate(False, "has no state")

    se.symbols = copy.deepcopy(symbols_original)
    se.symbols["1"] = "goodbye.jpg!!!"
    validate(False, "Could not load image")



import json
import itertools
from collections import OrderedDict
import numpy as np
from qtpy.QtGui import QColor
from qtpy.QtCore import Slot, Property, Qt
from .baseplot import BasePlot, NoDataError, BasePlotCurveItem
from .channel import PyDMChannel
from ..utilities import remove_protocol


DEFAULT_BUFFER_SIZE = 1200
MINIMUM_BUFFER_SIZE = 2


class ScatterPlotCurveItem(BasePlotCurveItem):
    _channels = ('x_channel', 'y_channel')

    def __init__(self, y_addr, x_addr, redraw_mode=None, bufferSizeChannelAddress=None, **kws):
        self.x_channel = None
        self.y_channel = None
        self.x_address = x_addr
        self.y_address = y_addr
        self.x_connected = False
        self.y_connected = False
        # If a name wasn't specified, use the addresses to make one.
        if kws.get('name') is None:
            if y_addr is None and x_addr is None:
                kws['name'] = ""
            else:
                y_name = remove_protocol(y_addr if y_addr is not None else "")
                x_name = remove_protocol(x_addr if x_addr is not None else "")
                kws['name'] = "{y} vs. {x}".format(y=y_name, x=x_name)
        self.redraw_mode = (redraw_mode if redraw_mode is not None
                            else self.REDRAW_ON_EITHER)
        self.bufferSizeChannel = None
        self.bufferSizeChannel_connected = False
        self._bufferSize = DEFAULT_BUFFER_SIZE
        self.data_buffer = np.zeros((2, self._bufferSize),
                                    order='f', dtype=float)
        self.points_accumulated = 0
        self.latest_x_value = None
        self.latest_y_value = None
        self.needs_new_x = True
        self.needs_new_y = True
        if 'symbol' not in kws.keys():
            kws['symbol'] = 'o'
        if 'lineStyle' not in kws.keys():
            kws['lineStyle'] = Qt.NoPen
        super(ScatterPlotCurveItem, self).__init__(**kws)
        self.bufferSizeChannelAddress = bufferSizeChannelAddress

    def to_dict(self):
        """
        Serialize this curve into a dictionary.

        Returns
        -------
        OrderedDict
            Representation with values for all properties
            needed to recreate this curve.
        """
        dic_ = OrderedDict([("y_channel", self.y_address),
                            ("x_channel", self.x_address)])
        dic_.update(super(ScatterPlotCurveItem, self).to_dict())
        dic_["redraw_mode"] = self.redraw_mode
        dic_['buffer_size'] = self.getBufferSize()
        dic_["bufferSizeChannelAddress"] = self.bufferSizeChannelAddress
        return dic_

    @property
    def x_address(self):
        """
        The address of the channel used to get the x axis data.

        Returns
        -------
        str
            The address of the channel used to get the x axis data.
        """
        if self.x_channel is None:
            return None
        return self.x_channel.address

    @x_address.setter
    def x_address(self, new_address):
        """
        The address of the channel used to get the x axis data.

        Parameters
        -------
        new_address: str
        """
        if new_address is None or len(str(new_address)) < 1:
            self.x_channel = None
            return
        self.x_channel = PyDMChannel(
            address=new_address,
            connection_slot=self.xConnectionStateChanged,
            value_slot=self.receiveXValue)

    @property
    def y_address(self):
        """
        The address of the channel used to get the y axis data.

        Returns
        -------
        str
            The address of the channel used to get the y axis data.
        """
        if self.y_channel is None:
            return None
        return self.y_channel.address

    @y_address.setter
    def y_address(self, new_address):
        """
        The address of the channel used to get the y axis data.

        Parameters
        ----------
        new_address: str
        """
        if new_address is None or len(str(new_address)) < 1:
            self.y_channel = None
            return
        self.y_channel = PyDMChannel(
            address=new_address,
            connection_slot=self.yConnectionStateChanged,
            value_slot=self.receiveYValue)

    @Slot(bool)
    def xConnectionStateChanged(self, connected):
        self.x_connected = connected

    @Slot(bool)
    def yConnectionStateChanged(self, connected):
        self.y_connected = connected

    @Slot(int)
    @Slot(float)
    def receiveXValue(self, new_x):
        """
        Handler for new x data.  This method is usually called by a PyDMChannel
        when it updates.  You can call this yourself to inject data into the curve.

        Parameters
        ----------
        new_x: numpy.ndarray
            A new array values for the X axis.
        """
        if new_x is None:
            return
        self.latest_x_value = new_x
        self.needs_new_x = False
        self.update_buffer()

    @Slot(int)
    @Slot(float)
    def receiveYValue(self, new_y):
        """
        Handler for new y data.  This method is usually called by a PyDMChannel
        when it updates.  You can call this yourself to inject data into the curve.

        Parameters
        ----------
        new_y: numpy.ndarray
            A new array values for the Y axis.
        """
        if new_y is None:
            return
        self.latest_y_value = new_y
        self.needs_new_y = False
        self.update_buffer()

    def update_buffer(self):
        """
        This is called whenever new data is received for X or Y.
        Based on the value of the redraw_mode attribute, it decides whether
        we are ready to shift the data buffer by one and add the latest data.
        """
        # If we haven't gotten values for X and Y yet, can't redraw.
        if self.latest_y_value is None or self.latest_x_value is None:
            return

        if self.redraw_mode == self.REDRAW_ON_EITHER:
            # no matter which channel updates, add a pair with the two most
            # recent values
            pass
        elif self.redraw_mode == self.REDRAW_ON_X:
            # If we only redraw when X updates, make sure new X data has
            # arrived since the last time we drew the plot.
            if self.needs_new_x:
                return
        elif self.redraw_mode == self.REDRAW_ON_Y:
            # If we only redraw when Y updates, make sure new Y data has
            # arrived since the last time we drew the plot.
            if self.needs_new_y:
                return
        elif self.redraw_mode == self.REDRAW_ON_BOTH:
            # Make sure both X and Y have received new data since the last
            # time we drew the plot.
            if self.needs_new_y or self.needs_new_x:
                return
        # If you get this far, we are OK to add the latest data to the buffer.
        self.data_buffer = np.roll(self.data_buffer, -1)
        self.data_buffer[0, -1] = self.latest_x_value
        self.data_buffer[1, -1] = self.latest_y_value
        if self.points_accumulated < self._bufferSize:
            self.points_accumulated = self.points_accumulated + 1
        self.data_changed.emit()

    def initialize_buffer(self):
        self.points_accumulated = 0
        self.data_buffer = np.zeros((2, self._bufferSize),
                                    order='f', dtype=float)

    def getBufferSize(self):
        return int(self._bufferSize)

    def setBufferSize(self, value):
        if self._bufferSize != int(value):
            self._bufferSize = max(int(value), MINIMUM_BUFFER_SIZE)
            self.initialize_buffer()

    def resetBufferSize(self):
        if self._bufferSize != DEFAULT_BUFFER_SIZE:
            self._bufferSize = DEFAULT_BUFFER_SIZE
            self.initialize_buffer()

    @property
    def bufferSizeChannelAddress(self):
        """
        The address of the channel used to get the buffer size.

        Returns
        -------
        str
            The address of the channel used to get the buffer size.
        """
        if self.bufferSizeChannel is None:
            return None
        return self.bufferSizeChannel.address

    @bufferSizeChannelAddress.setter
    def bufferSizeChannelAddress(self, new_address):
        """
        The address of the channel used to get the buffer size.

        Parameters
        ----------
        new_address: str
        """
        if new_address is None or len(str(new_address)) < 1:
            self.bufferSizeChannel = None
            return
        if self.bufferSizeChannel is not None and self.bufferSizeChannel.address == new_address:
            return
        self.bufferSizeChannel = PyDMChannel(
            address=new_address,
            connection_slot=self.bufferSizeConnectionStateChanged,
            value_slot=self.bufferSizeChannelValueReceiver)
        self.bufferSizeChannel.connect()

    @Slot(bool)
    def bufferSizeConnectionStateChanged(self, connected):
        self.bufferSizeChannel_connected = connected

    @Slot(int)
    def bufferSizeChannelValueReceiver(self, value):
        """
        Handler for change in buffer size.  This method is usually called by a PyDMChannel
        when it updates.  You can call this yourself to set or change the buffer size.

        Parameters
        ----------
        value: int
            A new value for the buffer size.
        """
        if value is None:
            return
        self.setBufferSize(value)
        self.update_buffer()

    def redrawCurve(self):
        """
        Called by the curve's parent plot whenever the curve needs to be
        re-drawn with new data.
        """
        self.setData(x=self.data_buffer[0, -self.points_accumulated:].astype(float),
                     y=self.data_buffer[1, -self.points_accumulated:].astype(float))
        self.needs_new_x = True
        self.needs_new_y = True

    def limits(self):
        """
        Get the limits of the data for this curve.

        Returns
        -------
        tuple
            A nested tuple of limits: ((xmin, xmax), (ymin, ymax))
        """
        if self.points_accumulated == 0:
            raise NoDataError("Curve has no data, cannot determine limits.")
        x_data = self.data_buffer[0, -self.points_accumulated:]
        y_data = self.data_buffer[1, -self.points_accumulated:]
        return ((float(np.amin(x_data)), float(np.amax(x_data))),
                (float(np.amin(y_data)), float(np.amax(y_data))))

    def channels(self):
        return [self.y_channel, self.x_channel]


class PyDMScatterPlot(BasePlot):
    """
    PyDMScatterPlot is a widget to plot one scalar value against another.
    Multiple scalar pairs can be plotted on the same plot.  Each pair has
    a buffer which stores previous values.  All values in the buffer are
    drawn.  The buffer size for each pair is user configurable.

    Parameters
    ----------
    parent : optional
        The parent of this widget.
    init_x_channels: optional
        init_x_channels can be a string with the address for a channel,
        or a list of strings, each containing an address for a channel.
        If not specified, y-axis waveforms will be plotted against their
        indices.  If a list is specified for both init_x_channels and
        init_y_channels, they both must have the same length.
        If a single x channel was specified, and a list of y channels are
        specified, all y channels will be plotted against the same x channel.
    init_y_channels: optional
        init_y_channels can be a string with the address for a channel,
        or a list of strings, each containing an address for a channel.
        If a list is specified for both init_x_channels and init_y_channels,
        they both must have the same length.
        If a single x channel was specified, and a list of y channels are
        specified, all y channels will be plotted against the same x channel.
    background: optional
        The background color for the plot. Accepts any arguments that
        pyqtgraph.mkColor will accept.
    """
    def __init__(self, parent=None, init_x_channels=[], init_y_channels=[],
                 background='default'):
        super(PyDMScatterPlot, self).__init__(parent, background)
        # If the user supplies a single string instead of a list,
        # wrap it in a list.
        if isinstance(init_x_channels, str):
            init_x_channels = [init_x_channels]
        if isinstance(init_y_channels, str):
            init_y_channels = [init_y_channels]
        if len(init_x_channels) == 0:
            init_x_channels = list(itertools.repeat(None,
                                                    len(init_y_channels)))
        if len(init_x_channels) != len(init_y_channels):
            raise ValueError("If lists are provided for both X and Y " +
                             "channels, they must be the same length.")
        # self.channel_pairs is an ordered dictionary that is keyed on a
        # (x_channel, y_channel) tuple, with ScatterPlotCurveItem values.
        # It gets populated in self.addChannel().
        self.channel_pairs = OrderedDict()
        init_channel_pairs = zip(init_x_channels, init_y_channels)
        for (x_chan, y_chan) in init_channel_pairs:
            self.addChannel(y_channel=y_chan, x_channel=x_chan)
        self._needs_redraw = True

    def initialize_for_designer(self):
        # If we are in Qt Designer, don't update the plot continuously.
        # This function gets called by PyDMTimePlot's designer plugin.
        pass

    def addChannel(self, y_channel=None, x_channel=None, name=None,
                   color=None, lineStyle=None, lineWidth=None,
                   symbol=None, symbolSize=None, redraw_mode=None,
                   buffer_size=None, yAxisName=None, bufferSizeChannelAddress=None):
        """
        Add a new curve to the plot.  In addition to the arguments below,
        all other keyword arguments are passed to the underlying
        pyqtgraph.PlotDataItem used to draw the curve.

        Parameters
        ----------
        y_channel: str
            The address for the y channel for the curve.
        x_channel: str, optional
            The address for the x channel for the curve.
        name: str, optional
            A name for this curve.  The name will be used in the plot legend.
        color: str or QColor, optional
            A color for the line of the curve.  If not specified, the plot will
            automatically assign a unique color from a set of default colors.
        lineStyle: int, optional
            Style of the line connecting the data points.
            0 means no line (scatter plot).
        lineWidth: int, optional
            Width of the line connecting the data points.
        redraw_mode: int, optional
            ScatterPlotCurveItem.REDRAW_ON_EITHER: (Default)
                Redraw after either X or Y receives new data.
            ScatterPlotCurveItem.REDRAW_ON_X:
                Redraw after X receives new data.
            ScatterPlotCurveItem.REDRAW_ON_Y:
                Redraw after Y receives new data.
            ScatterPlotCurveItem.REDRAW_ON_BOTH:
                Redraw after both X and Y receive new data.
        buffer_size: int, optional
            number of points to keep in the buffer.
        bufferSizeChannelAddress : str, optional
            The name of a channel that defines the buffer size (int).
        symbol: str or None, optional
            Which symbol to use to represent the data.
        symbol: int, optional
            Size of the symbol.
        yAxisName : str, optional
            The name of the y axis to associate with this curve. Will be created if it
            doesn't yet exist
        """
        plot_opts = {}
        plot_opts['symbol'] = symbol
        if symbolSize is not None:
            plot_opts['symbolSize'] = symbolSize
        if lineStyle is not None:
            plot_opts['lineStyle'] = lineStyle
        if lineWidth is not None:
            plot_opts['lineWidth'] = lineWidth
        if redraw_mode is not None:
            plot_opts['redraw_mode'] = redraw_mode
        curve = ScatterPlotCurveItem(y_addr=y_channel,
                                     x_addr=x_channel,
                                     name=name,
                                     color=color,
                                     yAxisName=yAxisName,
                                     bufferSizeChannelAddress=bufferSizeChannelAddress,
                                     **plot_opts)
        if buffer_size is not None:
            curve.setBufferSize(buffer_size)
        self.channel_pairs[(x_channel, y_channel)] = curve
        self.addCurve(curve, curve_color=color, y_axis_name=yAxisName)
        curve.data_changed.connect(self.set_needs_redraw)

    def removeChannel(self, curve):
        """
        Remove a curve from the plot.

        Parameters
        ----------
        curve: ScatterPlotCurveItem
            The curve to remove.
        """
        self.removeCurve(curve)

    def removeChannelAtIndex(self, index):
        """
        Remove a curve from the plot, given an index
        for a curve.

        Parameters
        ----------
        index: int
            Index for the curve to remove.
        """
        curve = self._curves[index]
        self.removeChannel(curve)

    @Slot()
    def set_needs_redraw(self):
        self._needs_redraw = True

    @Slot()
    def redrawPlot(self):
        """
        Request a redraw from each curve in the plot.
        Called by curves when they get new data.
        """
        if not self._needs_redraw:
            return
        for curve in self._curves:
            curve.redrawCurve()
        self._needs_redraw = False

    def clearCurves(self):
        """
        Remove all curves from the plot.
        """
        super(PyDMScatterPlot, self).clear()

    def getCurves(self):
        """
        Get a list of json representations for each curve.
        """
        return [json.dumps(curve.to_dict()) for curve in self._curves]

    def setCurves(self, new_list):
        """
        Replace all existing curves with new ones.  This function
        is mostly used as a way to load curves from a .ui file, and
        almost all users will want to add curves through addChannel,
        not this method.

        Parameters
        ----------
        new_list: list
            A list of json strings representing each curve in the plot.
        """
        try:
            new_list = [json.loads(str(i)) for i in new_list]
        except ValueError as e:
            print("Error parsing curve json data: {}".format(e))
            return
        self.clearCurves()
        for d in new_list:
            color = d.get('color')
            if color:
                color = QColor(color)
            self.addChannel(y_channel=d['y_channel'], x_channel=d['x_channel'],
                            name=d.get('name'), color=color,
                            lineStyle=d.get('lineStyle'),
                            lineWidth=d.get('lineWidth'),
                            symbol=d.get('symbol'),
                            symbolSize=d.get('symbolSize'),
                            redraw_mode=d.get('redraw_mode'),
                            buffer_size=d.get('buffer_size'),
                            bufferSizeChannelAddress=d.get('bufferSizeChannelAddress'),
                            yAxisName=d.get('yAxisName'))

    curves = Property("QStringList", getCurves, setCurves, designable=False)

    def channels(self):
        """
        Returns the list of channels used by all curves in the plot.

        Returns
        -------
        list
        """
        chans = []
        chans.extend([curve.y_channel for curve in self._curves])
        chans.extend([curve.x_channel for curve in self._curves])
        chans.extend([curve.bufferSizeChannel
                      for curve in self._curves
                      if curve.bufferSizeChannel is not None])
        return chans

    # The methods for autoRangeX, minXRange, maxXRange, autoRangeY, minYRange,
    # and maxYRange are all defined in BasePlot, but we don't expose them as
    # properties there, because not all plot subclasses necessarily want
    # them to be user-configurable in Designer.
    autoRangeX = Property(bool, BasePlot.getAutoRangeX,
                          BasePlot.setAutoRangeX, BasePlot.resetAutoRangeX,
                          doc="""
Whether or not the X-axis automatically rescales to fit the data.
If true, the values in minXRange and maxXRange are ignored.""")

    minXRange = Property(float, BasePlot.getMinXRange,
                         BasePlot.setMinXRange, doc="""
Minimum X-axis value visible on the plot.""")

    maxXRange = Property(float, BasePlot.getMaxXRange,
                         BasePlot.setMaxXRange, doc="""
Maximum X-axis value visible on the plot.""")

    autoRangeY = Property(bool, BasePlot.getAutoRangeY,
                          BasePlot.setAutoRangeY, BasePlot.resetAutoRangeY,
                          doc="""
Whether or not the Y-axis automatically rescales to fit the data.
If true, the values in minYRange and maxYRange are ignored.""")

    minYRange = Property(float, BasePlot.getMinYRange,
                         BasePlot.setMinYRange, doc="""
Minimum Y-axis value visible on the plot.""")

    maxYRange = Property(float, BasePlot.getMaxYRange,
                         BasePlot.setMaxYRange, doc="""
Maximum Y-axis value visible on the plot.""")



"""
External tool tests
"""

import pathlib
from typing import Any, Set

import pytest

from .. import tools
from ..application import PyDMApplication

EXAMPLE_PATH = pathlib.Path(__file__).parents[2] / "examples"
EXAMPLE_EXT_TOOL_PATH = EXAMPLE_PATH / "external_tool"

examples_required = pytest.mark.skipif(
    not EXAMPLE_PATH.exists(),
    reason="Not a source checkout - no examples available"
)


class ValidTool(tools.ExternalTool):
    ...


class InvalidTool:
    ...


@pytest.mark.parametrize(
    "cls, valid",
    [
        pytest.param(ValidTool, True, id="valid"),
        pytest.param(InvalidTool, False, id="invalid"),
        pytest.param(None, False, id="invalid-none"),
    ]
)
def test_valid_external_tool(cls: Any, valid: bool):
    assert tools._is_valid_external_tool_class(cls) is valid


@examples_required
@pytest.mark.parametrize(
    "source_file, expected_tools",
    [
        pytest.param(
            EXAMPLE_EXT_TOOL_PATH / "dummy_tool.py",
            {"DummyTool", "DummyTool3"},
            id="dummy_tool",
        ),
        pytest.param(
            EXAMPLE_EXT_TOOL_PATH / "lookup_path" / "new_tool.py",
            {"DummyTool2"},
            id="new_tool",
        ),
        pytest.param(
            EXAMPLE_EXT_TOOL_PATH / "lookup_path" / "root_tool.py",
            {"RootTool"},
            id="root_tool",
        ),
    ]
)
def test_tools_from_source(
    source_file: pathlib.Path,
    expected_tools: Set[str],
    qapp: PyDMApplication
):
    assert source_file.exists()
    loaded_tools = [
        tool.__class__.__name__
        for tool in tools._get_tools_from_source(str(source_file))
    ]
    assert set(loaded_tools) == expected_tools


def test_smoke_get_entrypoint_tools(qapp):
    list(tools.get_entrypoint_tools())


def test_smoke_get_tools_from_path(qapp):
    list(tools.get_tools_from_path())


def test_smoke_load_external_tools(qapp):
    tools.load_external_tools()



from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent
from qtpy.QtWidgets import QAction, QFrame, QApplication, QLabel, QMenu, QVBoxLayout, QWidget
from qtpy.QtCore import QPoint, Qt, QSize, Property, QTimer

import copy
import os.path
import logging
from .base import PyDMPrimitiveWidget
from .baseplot import BasePlot
from ..utilities import (is_pydm_app, establish_widget_connections,
                         close_widget_connections, macro, is_qt_designer,
                         find_file)
from ..display import (load_file, ScreenTarget)

logger = logging.getLogger(__name__)

_embeddedDisplayRuleProperties = {'Filename': ['filename', str]}

class PyDMEmbeddedDisplay(QFrame, PyDMPrimitiveWidget, new_properties=_embeddedDisplayRuleProperties):
    """
    A QFrame capable of rendering a PyDM Display

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label

    """

    def __init__(self, parent=None):
        QFrame.__init__(self, parent)
        PyDMPrimitiveWidget.__init__(self)
        self.app = QApplication.instance()
        self._filename = None
        self._macros = None
        self._embedded_widget = None
        self._disconnect_when_hidden = True
        self._is_connected = False
        self._only_load_when_shown = True
        self._needs_load = True
        self._load_error_timer = None
        self._load_error = None
        self._follow_symlinks = False
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)
        self.open_in_new_window_action = QAction('Open in New Window', self)
        self.open_in_new_window_action.triggered.connect(self.open_display_in_new_window)

        self.layout = QVBoxLayout(self)
        self.err_label = QLabel(self)
        self.err_label.setAlignment(Qt.AlignHCenter)
        self.layout.addWidget(self.err_label)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.err_label.hide()

    def init_for_designer(self):
        self.setFrameShape(QFrame.Box)

    def sizePolicy(self):
        """
        This holds the sizePolicy for the widget.

        Returns
        -------
        QSize
        """
        if self._embedded_widget is not None:
            return self._embedded_widget.sizePolicy()
        return super().sizePolicy()

    def sizeHint(self):
        """
        This holds the recommended size for the widget.

        Returns
        -------
        QSize
        """
        if self._embedded_widget is not None:
            return self._embedded_widget.sizeHint()
        return QSize(100, 100)

    def minimumSizeHint(self):
        """
        This holds the recommended minimum size for the widget.

        Returns
        -------
        QSize
        """
        if self._embedded_widget is not None:
            return self._embedded_widget.minimumSizeHint()
        # This is totally arbitrary, I just want *some* visible nonzero size
        return QSize(100, 100)

    @Property(str)
    def macros(self):
        """
        JSON-formatted string containing macro variables to pass to the embedded file.

        Returns
        -------
        str
        """
        if self._macros is None:
            return ""
        return self._macros

    @macros.setter
    def macros(self, new_macros):
        """
        JSON-formatted string containing macro variables to pass to the embedded file.

        .. warning::
        If the macros property is not defined before the filename property,
        The widget will not have any macros defined when it loads the embedded file.
        This behavior will be fixed soon.

        Parameters
        ----------
        new_macros : str
        """
        new_macros = str(new_macros)
        if new_macros != self._macros:
            self._macros = new_macros
            self._needs_load = True
            self.load_if_needed()

    @Property(str)
    def filename(self):
        """
        Filename of the display to embed.

        Returns
        -------
        str
        """
        if self._filename is None:
            return ""
        return self._filename

    @filename.setter
    def filename(self, filename):
        """
        Filename of the display to embed.

        Parameters
        ----------
        filename : str
        """
        filename = str(filename)
        if filename != self._filename:
            self._filename = filename
            self._needs_load = True
            if is_qt_designer():
                if self._load_error_timer:
                    # Kill the timer here. If new filename still causes the problem, it will be restarted
                    self._load_error_timer.stop()
                    self._load_error_timer = None
                self.clear_error_text()
            self.load_if_needed()

    def set_macros_and_filename(self, new_filename, new_macros):
        """
        A method to change both macros and the filename of an embedded display.
        the method takes in a Filename of the display to embed and a
        JSON-formatted string containing macro variables to pass to the embedded file.

        Parameters
        ----------
        new_macros : str
        new_filename : str
        """
        new_macros = str(new_macros)
        if new_macros != self._macros:
            self._macros = new_macros

        self.filename = new_filename

    def parsed_macros(self):
        """
        Dictionary containing the key value pair for each macro specified.

        Returns
        --------
        dict
        """
        parent_display = self.find_parent_display()
        parent_macros = {}
        if parent_display:
            parent_macros = copy.copy(parent_display.macros())
        widget_macros = macro.parse_macro_string(self.macros)
        parent_macros.update(widget_macros)
        return parent_macros

    def load_if_needed(self):
        if self._needs_load and (
                not self._only_load_when_shown or self.isVisible() or is_qt_designer()):
            self.embedded_widget = self.open_file()

    def open_file(self, force=False):
        """
        Opens the widget specified in the widget's filename property.

        Returns
        -------
        display : QWidget
        """
        if (not force) and (not self._needs_load):
            return
            
        if not self.filename:
            return

        try:
            parent_display = self.find_parent_display()
            base_path = ""
            if parent_display:
                parent_file_path = parent_display.loaded_file()
                if self._follow_symlinks:
                    parent_file_path = os.path.realpath(parent_file_path)
                base_path = os.path.dirname(parent_file_path)

            fname = find_file(self.filename, base_path=base_path, raise_if_not_found=True)
            w = load_file(fname, macros=self.parsed_macros(), target=None)
            self._needs_load = False
            self.clear_error_text()
            return w
        except Exception as e:
            self._load_error = e
            if self._load_error_timer:
                self._load_error_timer.stop()
            self._load_error_timer = QTimer(self)
            self._load_error_timer.setSingleShot(True)
            self._load_error_timer.setTimerType(Qt.VeryCoarseTimer)
            self._load_error_timer.timeout.connect(self._display_designer_load_error)
            self._load_error_timer.start(1000)
        return None

    def clear_error_text(self):
        if self._load_error_timer:
            self._load_error_timer.stop()
        self.err_label.clear()
        self.err_label.hide()

    def display_error_text(self, e):
        self.err_label.setText(
            "Could not open {filename}.\nError: {err}".format(
                filename=self._filename, err=e))
        self.err_label.show()

    @property
    def embedded_widget(self):
        """
        The embedded widget being displayed.

        Returns
        -------
        QWidget
        """
        return self._embedded_widget

    @embedded_widget.setter
    def embedded_widget(self, new_widget):
        """
        Defines the embedded widget to display inside the QFrame

        Parameters
        ----------
        new_widget : QWidget
        """
        should_reconnect = False
        if new_widget is self._embedded_widget:
            return
        if self._embedded_widget is not None:
            self.layout.removeWidget(self._embedded_widget)
            self._embedded_widget.deleteLater()
            self._embedded_widget = None
        if new_widget is not None:
            self._embedded_widget = new_widget
            self._embedded_widget.setParent(self)
            self.layout.addWidget(self._embedded_widget)
            self.err_label.hide()
            self._embedded_widget.show()
            self._is_connected = True

    def connect(self):
        """
        Establish the connection between the embedded widget and
        the channels associated with it.
        """
        if self._is_connected or self.embedded_widget is None:
            return
        establish_widget_connections(self.embedded_widget)
        self._is_connected = True

    def disconnect(self):
        """
        Disconnects the embedded widget from the channels
        associated with it.
        """
        if not self._is_connected or self.embedded_widget is None:
            return
        close_widget_connections(self.embedded_widget)
        self._is_connected = False

    @Property(bool)
    def loadWhenShown(self):
        """
        If True, only load and display the file once the
        PyDMEmbeddedDisplayWidget is visible on screen.  This is very useful
        if you have many different PyDMEmbeddedWidgets in different tabs of a
        QTabBar or PyDMTabBar: only the tab that the user is looking at will
        be loaded, which can greatly speed up the launch time of a display.
        
        If this property is changed from 'True' to 'False', and the file has
        not been loaded yet, it will be loaded immediately.
        
        Returns
        -------
        bool
        """
        return self._only_load_when_shown
        
    @loadWhenShown.setter
    def loadWhenShown(self, val):
        self._only_load_when_shown = val
        self.load_if_needed()

    @Property(bool)
    def disconnectWhenHidden(self):
        """
        Disconnect from PVs when this widget is not visible.

        Returns
        -------
        bool
        """
        return self._disconnect_when_hidden

    @disconnectWhenHidden.setter
    def disconnectWhenHidden(self, disconnect_when_hidden):
        """
        Disconnect from PVs when this widget is not visible.

        Parameters
        ----------
        disconnect_when_hidden : bool
        """
        self._disconnect_when_hidden = disconnect_when_hidden

    @Property(bool)
    def followSymlinks(self) -> bool:
        """
        If True, any symlinks in the path to filename (including the base path of the parent display) will be followed, so that it
        will always use the canonical path. If False (default), the file will be searched without canonicalizing the path beforehand.

        Note that it will not work on Windows if you're using a Python version prior to 3.8.

        Returns
        -------
        bool
        """
        return self._follow_symlinks

    @followSymlinks.setter
    def followSymlinks(self, follow_symlinks: bool) -> None:
        """
        If True, any symlinks in the path to filename (including the base path of the parent display) will be followed, so that it
        will always use the canonical path. If False (default), the file will be searched using the non-canonical path.

        Note that it will not work on Windows if you're using a Python version prior to 3.8.

        Parameters
        ----------
        follow_symlinks : bool
        """
        self._follow_symlinks = follow_symlinks

    def showEvent(self, e):
        """
        Show events are sent to widgets that become visible on the screen.

        Parameters
        ----------
        event : QShowEvent
        """
        if self._only_load_when_shown:
            w = self.open_file()
            if w:
                self.embedded_widget = w
        if self.disconnectWhenHidden:
            self.connect()

    def hideEvent(self, e):
        """
        Hide events are sent to widgets that become invisible on the screen.

        Parameters
        ----------
        event : QHideEvent
        """
        if self.disconnectWhenHidden:
            self.disconnect()

    def _display_designer_load_error(self):
        self._load_error_timer = None
        logger.exception("Exception while opening embedded display file.", exc_info=self._load_error)
        if self._load_error:
            self.display_error_text(self._load_error)

    def open_display_in_new_window(self) -> None:
        """ Open the embedded display in a new window """
        if not self.filename:
            return
        file_path = find_file(self.filename, base_path='', raise_if_not_found=True)
        macros = self.parsed_macros()

        if is_pydm_app():
            load_file(file_path, macros=macros)
        else:
            w = load_file(file_path, macros=macros, target=ScreenTarget.DIALOG)

    def create_context_menu(self, pos: QPoint) -> QMenu:
        """ Create the right-click context menu for this embedded widget based on the location of the mouse click """
        if self._embedded_widget is None:
            return

        menu = None
        # Plot widgets use their own custom event handling, so we check to see if they were
        # clicked on here. If so, just re-use the context menu they already have built. (Not
        # specifically checking for these would clobber their context menus)
        plot_widgets = self.findChildren(BasePlot)
        for plot in plot_widgets:
            try:
                if plot.geometry().contains(pos):
                    menu = plot.getViewBox().getMenu(None)
                    if menu is not None:  # Need to add sub-menus still
                        # Mock up an event that the pyqtgraph api requires. Just want to create it without any
                        # initialization of attributes (would require more unnecessary object creations)
                        accept_event = object.__new__(MouseClickEvent)
                        accept_event.accepted = True
                        accept_event.acceptedItem = plot.getViewBox()
                        menu = plot.getViewBox().scene().addParentContextMenus(plot.getViewBox(), menu, accept_event)
            except AttributeError:
                pass

        # Otherwise check if a menu already exists, and create a new one if not
        if menu is None:
            try:
                menu = self._embedded_widget.context_menu()
            except AttributeError:
                menu = QMenu(self)

        if len(menu.findChildren(QAction)) > 0:
            menu.addSeparator()
        menu.addAction(self.open_in_new_window_action)
        return menu

    def show_context_menu(self, pos: QPoint) -> None:
        """ Display the right-click context menu for this embedded widget at the location of the mouse click """
        menu = self.create_context_menu(pos)
        if menu is not None:
            menu.exec_(self.mapToGlobal(pos))
